#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES).
#
# Copyright (c) 2018-2024 by the software owners: The Regents of the
# University of California, through Lawrence Berkeley National Laboratory,
# National Technology & Engineering Solutions of Sandia, LLC, Carnegie Mellon
# University, West Virginia University Research Corporation, et al.
# All rights reserved.  Please see the files COPYRIGHT.md and LICENSE.md
# for full copyright and license information.
#################################################################################
"""
Unit operation model for a steam heater applicable to platen superheater
and roof superheater, model main equations:

* Heat is given by fire-side boiler model
* Calculate pressure change due to friction
* Calculate slag layer wall temperature
* Consider a layer of metal and a layer of slag

"""
# TODO: Missing docstrings
# pylint: disable=missing-function-docstring

# Import Pyomo libraries
from pyomo.common.config import ConfigBlock, ConfigValue, In, Bool
from pyomo.environ import value, Var, Param, asin, cos, Reference
from pyomo.core.expr import Expr_if
from pyomo.dae import DerivativeVar

# Import IDAES cores
from idaes.core import (
    ControlVolume0DBlock,
    declare_process_block_class,
    MaterialBalanceType,
    EnergyBalanceType,
    MomentumBalanceType,
    UnitModelBlockData,
    useDefault,
)
from idaes.core.util.config import is_physical_parameter_block
import idaes.logger as idaeslog
import idaes.core.util.scaling as iscale
from idaes.core.solvers import get_solver
from idaes.core.util.constants import Constants as const


__author__ = "Boiler Subsystem Team (J. Ma, M. Zamarripa)"
__version__ = "2.0.0"


@declare_process_block_class("SteamHeater")
class SteamHeaterData(UnitModelBlockData):
    """
    WaterwallSection Unit Class
    """

    CONFIG = UnitModelBlockData.CONFIG()

    CONFIG.declare(
        "material_balance_type",
        ConfigValue(
            default=MaterialBalanceType.componentPhase,
            domain=In(MaterialBalanceType),
            description="Material balance construction flag",
            doc="""Indicates what type of material balance should be constructed,
**default** - MaterialBalanceType.componentPhase.
**Valid values:** {
**MaterialBalanceType.none** - exclude material balances,
**MaterialBalanceType.componentPhase** - use phase component balances,
**MaterialBalanceType.componentTotal** - use total component balances,
**MaterialBalanceType.elementTotal** - use total element balances,
**MaterialBalanceType.total** - use total material balance.}""",
        ),
    )
    CONFIG.declare(
        "energy_balance_type",
        ConfigValue(
            default=EnergyBalanceType.enthalpyTotal,
            domain=In(EnergyBalanceType),
            description="Energy balance construction flag",
            doc="""Indicates what type of energy balance should be constructed,
**default** - EnergyBalanceType.enthalpyTotal.
**Valid values:** {
**EnergyBalanceType.none** - exclude energy balances,
**EnergyBalanceType.enthalpyTotal** - single ethalpy balance for material,
**EnergyBalanceType.enthalpyPhase** - ethalpy balances for each phase,
**EnergyBalanceType.energyTotal** - single energy balance for material,
**EnergyBalanceType.energyPhase** - energy balances for each phase.}""",
        ),
    )
    CONFIG.declare(
        "momentum_balance_type",
        ConfigValue(
            default=MomentumBalanceType.pressureTotal,
            domain=In(MomentumBalanceType),
            description="Momentum balance construction flag",
            doc="""Indicates what type of momentum balance should be constructed,
**default** - MomentumBalanceType.pressureTotal.
**Valid values:** {
**MomentumBalanceType.none** - exclude momentum balances,
**MomentumBalanceType.pressureTotal** - single pressure balance for material,
**MomentumBalanceType.pressurePhase** - pressure balances for each phase,
**MomentumBalanceType.momentumTotal** - single momentum balance for material,
**MomentumBalanceType.momentumPhase** - momentum balances for each phase.}""",
        ),
    )
    CONFIG.declare(
        "has_heat_transfer",
        ConfigValue(
            default=False,
            domain=Bool,
            description="Heat transfer term construction flag",
            doc="""Indicates whether terms for heat transfer should be constructed,
**default** - False.
**Valid values:** {
**True** - include heat transfer terms,
**False** - exclude heat transfer terms.}""",
        ),
    )
    CONFIG.declare(
        "has_pressure_change",
        ConfigValue(
            default=False,
            domain=Bool,
            description="Pressure change term construction flag",
            doc="""Indicates whether terms for pressure change should be
constructed,
**default** - False.
**Valid values:** {
**True** - include pressure change terms,
**False** - exclude pressure change terms.}""",
        ),
    )
    CONFIG.declare(
        "property_package",
        ConfigValue(
            default=useDefault,
            domain=is_physical_parameter_block,
            description="Property package to use for control volume",
            doc="""Property parameter object used to define property calculations,
**default** - useDefault.
**Valid values:** {
**useDefault** - use default package from parent model or flowsheet,
**PhysicalParameterObject** - a PhysicalParameterBlock object.}""",
        ),
    )
    CONFIG.declare(
        "property_package_args",
        ConfigBlock(
            implicit=True,
            description="Arguments to use for constructing property packages",
            doc="""A ConfigBlock with arguments to be passed to a property block(s)
and used when constructing these,
**default** - None.
**Valid values:** {
see property package for documentation.}""",
        ),
    )
    CONFIG.declare(
        "single_side_only",
        ConfigValue(
            default=True,
            domain=Bool,
            description="Flag indicating the heat is from one side of tubes only",
            doc="""Indicates whether tubes are heated from one side only,
**default** - True.
**Valid values:** {
**True** - single side is heated such as roof,
**False** - both sides are heated such as platen superheater.}""",
        ),
    )

    def build(self):
        """
        Build control volume and ports
        """
        # Call UnitModel.build to setup dynamics
        super(SteamHeaterData, self).build()

        # Build Control Volume
        self.control_volume = ControlVolume0DBlock(
            dynamic=self.config.dynamic,
            has_holdup=self.config.has_holdup,
            property_package=self.config.property_package,
            property_package_args=self.config.property_package_args,
        )

        self.control_volume.add_geometry()

        self.control_volume.add_state_blocks(has_phase_equilibrium=False)

        self.control_volume.add_material_balances(
            balance_type=self.config.material_balance_type
        )

        self.control_volume.add_energy_balances(
            balance_type=self.config.energy_balance_type,
            has_heat_transfer=self.config.has_heat_transfer,
        )

        self.control_volume.add_momentum_balances(
            balance_type=self.config.momentum_balance_type, has_pressure_change=True
        )

        # Add Ports
        self.add_inlet_port()
        self.add_outlet_port()

        # Add object references
        self.volume = Reference(self.control_volume.volume)

        # Set references to balance terms at unit level
        if (
            self.config.has_heat_transfer is True
            and self.config.energy_balance_type != EnergyBalanceType.none
        ):
            self.heat_duty = Reference(self.control_volume.heat)
        if (
            self.config.has_pressure_change is True
            and self.config.momentum_balance_type != "none"
        ):
            self.deltaP = Reference(self.control_volume.deltaP)

        # Set Unit Geometry and Holdup Volume
        self._set_geometry()

        # Construct performance equations
        self._make_performance()

    def _set_geometry(self):
        """
        Define the geometry of the unit as necessary, and link to holdup volume
        """
        # Number of tubes that steam flows through
        self.number_tubes = Var(initialize=4, doc="Number of tubes")

        # Average length of tubes that steam flows through from inlet to outlet
        self.tube_length = Var(initialize=5.0, doc="length tube from inlet to outlet")

        # Inside diameter of tubes
        self.diameter_in = Var(initialize=0.05, doc="Inside diameter of tubes")

        # Inside radius of tube
        @self.Expression(doc="Inside radius of tube")
        def radius_in(b):
            return 0.5 * b.diameter_in

        # Total cross section area of fluid flow
        @self.Expression(doc="Cross section area of fluid")
        def area_cross_fluid_total(b):
            return 0.25 * const.pi * b.diameter_in**2 * b.number_tubes

        # Tube thickness
        self.tube_thickness = Var(initialize=0.005, doc="Thickness of tube")

        # Outside radius of tube
        @self.Expression(doc="Outside radius of tube")
        def radius_out(b):
            return b.radius_in + b.tube_thickness

        # Thickness of fin
        self.fin_thickness = Var(initialize=0.004, doc="Thickness of fin")

        # Length of fin
        self.fin_length = Var(initialize=0.005, doc="Length of fin")
        # Thickness of slag layer
        self.slag_thickness = Var(
            self.flowsheet().time, initialize=0.001, doc="thickness of slag layer"
        )

        @self.Expression(doc="Pitch of two neighboring tubes")
        def pitch(b):
            return b.fin_length + b.radius_out * 2.0

        # total projected area
        @self.Expression(doc="total projected area for heat transfer")
        def area_proj_total(b):
            if self.config.single_side_only:
                return b.tube_length * b.pitch * b.number_tubes
            else:
                return 2 * b.tube_length * b.pitch * b.number_tubes

        @self.Expression(doc="Angle at joint of tube and fin")
        def alpha_tube(b):
            return asin(0.5 * b.fin_thickness / b.radius_out)

        @self.Expression(
            self.flowsheet().time,
            doc="Angle at joint of tube " "and fin at outside slag layer",
        )
        def alpha_slag(b, t):
            return asin(
                (0.5 * b.fin_thickness + b.slag_thickness[t])
                / (b.radius_out + b.slag_thickness[t])
            )

        @self.Expression(doc="Perimeter of interface between slag and tube")
        def perimeter_if(b):
            if self.config.single_side_only:
                return (
                    (const.pi - 2 * b.alpha_tube) * b.radius_out
                    + b.pitch
                    - 2 * b.radius_out * cos(b.alpha_tube)
                )
            else:
                return 2 * (
                    (const.pi - 2 * b.alpha_tube) * b.radius_out
                    + b.pitch
                    - 2 * b.radius_out * cos(b.alpha_tube)
                )

        @self.Expression(doc="Perimeter on the inner tube side")
        def perimeter_ts(b):
            return const.pi * b.diameter_in

        @self.Expression(self.flowsheet().time, doc="Perimeter on the outer slag side")
        def perimeter_ss(b, t):
            if self.config.single_side_only:
                return (
                    (const.pi - 2 * b.alpha_slag[t])
                    * (b.radius_out + b.slag_thickness[t])
                    + b.pitch
                    - 2 * (b.radius_out + b.slag_thickness[t]) * cos(b.alpha_slag[t])
                )
            else:
                return 2 * (
                    (const.pi - 2 * b.alpha_slag[t])
                    * (b.radius_out + b.slag_thickness[t])
                    + b.pitch
                    - 2 * (b.radius_out + b.slag_thickness[t]) * cos(b.alpha_slag[t])
                )

        # Cross section area of tube and fin metal
        @self.Expression(doc="Cross section area of tube and fin metal")
        def area_cross_metal(b):
            return (
                const.pi * (b.radius_out**2 - b.radius_in**2)
                + b.fin_thickness * b.fin_length
            )

        # Cross section area of slag layer
        @self.Expression(
            self.flowsheet().time, doc="Cross section area of slag layer per tube"
        )
        def area_cross_slag(b, t):
            return b.perimeter_if * b.slag_thickness[t]

        # Volume constraint
        @self.Constraint(
            self.flowsheet().time, doc="waterwall fluid volume of all tubes"
        )
        def volume_eqn(b, t):
            return (
                b.volume[t]
                == 0.25 * const.pi * b.diameter_in**2 * b.tube_length * b.number_tubes
            )

    def _make_performance(self):
        """
        Define constraints which describe the behaviour of the unit model.
        """
        # Thermal conductivity of metal
        self.therm_cond_metal = Param(
            initialize=43.0, mutable=True, doc="Thermal conductivity of tube metal"
        )
        # Thermal conductivity of slag
        self.therm_cond_slag = Var(initialize=1.3, doc="Thermal conductivity of slag")
        # Heat capacity of metal
        self.cp_metal = Param(
            initialize=500.0, mutable=True, doc="Heat capacity of tube metal"
        )
        # Heat Capacity of slag
        self.cp_slag = Param(initialize=250, mutable=True, doc="Heat capacity of slag")
        # Density of metal
        self.dens_metal = Param(
            initialize=7800.0, mutable=True, doc="Density of tube metal"
        )
        # Density of slag
        self.dens_slag = Param(initialize=2550, mutable=True, doc="Density of slag")
        # Shape factor of tube metal conduction
        self.fshape_metal = Param(
            initialize=1.0, mutable=True, doc="Shape factor of tube metal conduction"
        )
        # Shape factor of slag conduction
        self.fshape_slag = Param(
            initialize=1.0, mutable=True, doc="Shape factor of slag conduction"
        )

        # Add performance variables
        # Heat from fire side boiler model
        self.heat_fireside = Var(
            self.flowsheet().time,
            initialize=1e7,
            doc="total heat from fire side model for the section",
        )
        # Tube boundary wall temperature
        self.temp_tube_boundary = Var(
            self.flowsheet().time,
            initialize=400.0,
            doc="Temperature of tube boundary wall",
        )
        # Tube center point wall temperature
        self.temp_tube_center = Var(
            self.flowsheet().time,
            initialize=450.0,
            doc="Temperature of tube center wall",
        )
        # Slag boundary wall temperature
        self.temp_slag_boundary = Var(
            self.flowsheet().time,
            initialize=600.0,
            doc="Temperature of slag boundary wall",
        )
        # Slag center point slag wall temperature
        self.temp_slag_center = Var(
            self.flowsheet().time,
            initialize=500.0,
            doc="Temperature of slag layer center point",
        )

        # Energy holdup for slag layer
        self.energy_holdup_slag = Var(
            self.flowsheet().time, initialize=1.0, doc="Energy holdup of slag layer"
        )

        # Energy holdup for metal (tube + fin)
        self.energy_holdup_metal = Var(
            self.flowsheet().time, initialize=1.0, doc="Energy holdup of metal"
        )

        # Energy accumulation for slag and metal
        if self.config.dynamic is True:
            self.energy_accumulation_slag = DerivativeVar(
                self.energy_holdup_slag,
                wrt=self.flowsheet().time,
                doc="Energy accumulation of slag layer",
            )
            self.energy_accumulation_metal = DerivativeVar(
                self.energy_holdup_metal,
                wrt=self.flowsheet().time,
                doc="Energy accumulation of tube and fin metal",
            )

        def energy_accumulation_term_slag(b, t):
            return b.energy_accumulation_slag[t] if b.config.dynamic else 0

        def energy_accumulation_term_metal(b, t):
            return b.energy_accumulation_metal[t] if b.config.dynamic else 0

        # Velocity of steam
        self.velocity = Var(
            self.flowsheet().time, initialize=3.0, doc="Velocity of steam"
        )

        # Reynolds number based on liquid only flow
        self.N_Re = Var(self.flowsheet().time, initialize=1.0e6, doc="Reynolds number")

        # Prandtl number of liquid phase
        self.N_Pr = Var(self.flowsheet().time, initialize=2.0, doc="Reynolds number")

        # Darcy friction factor
        self.friction_factor_darcy = Var(
            self.flowsheet().time, initialize=0.01, doc="Darcy friction factor"
        )

        # Convective heat transfer coefficient on tube side,
        # typically in range (1000, 5e5)
        self.hconv = Var(
            self.flowsheet().time,
            initialize=30000.0,
            doc="Convective heat transfer coefficient",
        )

        # Convective heat flux to fluid
        self.heat_flux_conv = Var(
            self.flowsheet().time, initialize=7e4, doc="Convective heat flux to fluid"
        )

        # Fire-side heat flux
        self.heat_flux_fireside = Var(
            self.flowsheet().time,
            initialize=100000.0,
            doc="Fireside heat flux to slag boundary",
        )

        # Slag-tube interface heat flux
        self.heat_flux_interface = Var(
            self.flowsheet().time,
            initialize=100000.0,
            doc="Slag-tube interface heat flux",
        )

        # Equation to calculate heat flux to slag boundary
        @self.Constraint(self.flowsheet().time, doc="heat flux at slag outer layer")
        def heat_flux_fireside_from_boiler_eqn(b, t):
            if self.config.single_side_only:
                return (
                    b.heat_flux_fireside[t] * b.area_proj_total * b.perimeter_ss[t]
                    == b.heat_fireside[t] * b.pitch
                )
            else:
                return (
                    b.heat_flux_fireside[t] * b.area_proj_total * b.perimeter_ss[t]
                    == b.heat_fireside[t] * b.pitch * 2.0
                )

        # Equation to calculate slag layer boundary temperature
        @self.Constraint(self.flowsheet().time, doc="slag layer boundary temperature")
        def slag_layer_boundary_temperature_eqn(b, t):
            return b.heat_flux_fireside[t] * 0.5 * b.slag_thickness[
                t
            ] == b.fshape_slag * b.therm_cond_slag * (
                b.temp_slag_boundary[t] - b.temp_slag_center[t]
            )

        # Equation to calculate heat flux at the slag-metal interface
        @self.Constraint(self.flowsheet().time, doc="heat flux at slag-tube interface")
        def heat_flux_interface_eqn(b, t):
            return (
                b.heat_flux_interface[t]
                * 0.5
                * (
                    b.slag_thickness[t] / b.therm_cond_slag / b.fshape_slag
                    + b.tube_thickness / b.therm_cond_metal / b.fshape_metal
                )
                == b.temp_slag_center[t] - b.temp_tube_center[t]
            )

        # Equation to calculate heat flux at tube boundary
        @self.Constraint(
            self.flowsheet().time, doc="convective heat flux at tube boundary"
        )
        def heat_flux_conv_eqn(b, t):
            return b.heat_flux_conv[t] == b.hconv[t] * (
                b.temp_tube_boundary[t]
                - (
                    b.control_volume.properties_in[t].temperature
                    + b.control_volume.properties_out[t].temperature
                )
                / 2.0
            )

        # Equation to calculate tube boundary wall temperature
        @self.Constraint(self.flowsheet().time, doc="tube boundary wall temperature")
        def temperature_tube_boundary_eqn(b, t):
            return b.heat_flux_conv[
                t
            ] * 0.5 * b.tube_thickness == b.fshape_metal * b.therm_cond_metal * (
                b.temp_tube_center[t] - b.temp_tube_boundary[t]
            )

        # Equation to calculate energy holdup for slag layer per tube length
        @self.Constraint(self.flowsheet().time, doc="energy holdup for slag layer")
        def energy_holdup_slag_eqn(b, t):
            return (
                b.energy_holdup_slag[t]
                == b.temp_slag_center[t]
                * b.cp_slag
                * b.dens_slag
                * b.area_cross_slag[t]
            )

        # Equation to calculate energy holdup for metal
        # (tube + fin) per tube length
        @self.Constraint(self.flowsheet().time, doc="energy holdup for metal")
        def energy_holdup_metal_eqn(b, t):
            return (
                b.energy_holdup_metal[t]
                == b.temp_tube_center[t]
                * b.cp_metal
                * b.dens_metal
                * b.area_cross_metal
            )

        # Energy balance for slag layer
        @self.Constraint(self.flowsheet().time, doc="energy balance for slag layer")
        def energy_balance_slag_eqn(b, t):
            return energy_accumulation_term_slag(b, t) == (
                b.heat_flux_fireside[t] * b.perimeter_ss[t]
                - b.heat_flux_interface[t] * b.perimeter_if
            )

        # Energy balance for metal
        @self.Constraint(self.flowsheet().time, doc="energy balance for metal")
        def energy_balance_metal_eqn(b, t):
            return energy_accumulation_term_metal(b, t) == (
                b.heat_flux_interface[t] * b.perimeter_if
                - b.heat_flux_conv[t] * b.perimeter_ts
            )

        # Expression to calculate slag/tube metal interface wall temperature
        @self.Expression(
            self.flowsheet().time, doc="Slag tube interface wall temperature"
        )
        def temp_interface(b, t):
            return (
                b.temp_tube_center[t]
                + b.heat_flux_interface[t]
                * 0.5
                * b.tube_thickness
                / b.therm_cond_metal
                / b.fshape_metal
            )

        # Equations for calculate pressure drop
        # and convective heat transfer coefficient
        # Equation for calculating velocity
        @self.Constraint(self.flowsheet().time, doc="Vecolity of fluid")
        def velocity_eqn(b, t):
            return (
                1e-3
                * b.velocity[t]
                * b.area_cross_fluid_total
                * b.control_volume.properties_in[t].dens_mol_phase["Vap"]
                == 1e-3 * b.control_volume.properties_in[t].flow_mol
            )

        # Equation for calculating Reynolds number if liquid only
        @self.Constraint(self.flowsheet().time, doc="Reynolds number")
        def Reynolds_number_eqn(b, t):
            return (
                b.N_Re[t] * b.control_volume.properties_in[t].visc_d_phase["Vap"]
                == b.diameter_in
                * b.velocity[t]
                * b.control_volume.properties_in[t].dens_mass
            )

        # Friction factor depending on laminar or turbulent flow
        @self.Constraint(self.flowsheet().time, doc="Darcy friction factor")
        def friction_factor_darcy_eqn(b, t):
            return (
                Expr_if(
                    b.N_Re[t] < 1187.384,
                    b.friction_factor_darcy[t] * b.N_Re[t] / 64.0,
                    b.friction_factor_darcy[t] * b.N_Re[t] ** 0.25 / 0.3164,
                )
                == 1.0
            )

        # Pressure change equation due to friction,
        # -1/2*density*velocity^2*fD/diameter*length
        @self.Constraint(self.flowsheet().time, doc="pressure change due to friction")
        def pressure_change_eqn(b, t):
            return (
                b.deltaP[t] * b.diameter_in
                == -0.5
                * b.control_volume.properties_in[t].dens_mass
                * b.velocity[t] ** 2
                * b.friction_factor_darcy[t]
                * b.tube_length
            )

        # Total heat added to control_volume
        @self.Constraint(
            self.flowsheet().time, doc="total heat added to fluid control_volume"
        )
        def heat_eqn(b, t):
            return (
                b.heat_duty[t]
                == b.number_tubes * b.heat_flux_conv[t] * b.tube_length * b.perimeter_ts
            )

        # Prandtl number of steam
        @self.Constraint(self.flowsheet().time, doc="Prandtl number")
        def N_Pr_eqn(b, t):
            return (
                b.N_Pr[t]
                * b.control_volume.properties_in[t].therm_cond_phase["Vap"]
                * b.control_volume.properties_in[t].mw
                == b.control_volume.properties_in[t].cp_mol_phase["Vap"]
                * b.control_volume.properties_in[t].visc_d_phase["Vap"]
            )

        # Forced convection heat transfer coefficient for liquid only
        @self.Constraint(
            self.flowsheet().time,
            doc="forced convection heat transfer" " coefficient for liquid only",
        )
        def hconv_eqn(b, t):
            return (
                b.hconv[t] * b.diameter_in
                == 0.023
                * b.N_Re[t] ** 0.8
                * b.N_Pr[t] ** 0.4
                * b.control_volume.properties_in[t].therm_cond_phase["Vap"]
            )

    def set_initial_condition(self):
        if self.config.dynamic is True:
            self.control_volume.material_accumulation[:, :, :].value = 0
            self.control_volume.energy_accumulation[:, :].value = 0
            self.energy_accumulation_slag[:].value = 0
            self.energy_accumulation_metal[:].value = 0
            self.control_volume.material_accumulation[0, :, :].fix(0)
            self.control_volume.energy_accumulation[0, :].fix(0)
            self.energy_accumulation_slag[0].fix(0)
            self.energy_accumulation_metal[0].fix(0)

    def initialize_build(
        blk, state_args=None, outlvl=idaeslog.NOTSET, solver=None, optarg=None
    ):
        """
        Waterwall section initialization routine.

        Keyword Arguments:
            state_args : a dict of arguments to be passed to the property
                           package(s) for the control_volume of the model to
                           provide an initial state for initialization
                           (see documentation of the specific property package)
                           (default = None).
            outlvl : sets output level of initialisation routine

                     * 0 = no output (default)
                     * 1 = return solver state for each step in routine
                     * 2 = return solver state for each step in subroutines
                     * 3 = include solver output information (tee=True)

            optarg : solver options dictionary object (default=None, use
                     default solver options)
            solver : str indicating which solver to use during
                     initialization (default = None, use default solver)

        Returns:
            None
        """
        init_log = idaeslog.getInitLogger(blk.name, outlvl, tag="unit")
        solve_log = idaeslog.getSolveLogger(blk.name, outlvl, tag="unit")

        # Create solver
        opt = get_solver(solver, optarg)

        flags = blk.control_volume.initialize(
            outlvl=outlvl, optarg=optarg, solver=solver, state_args=state_args
        )
        init_log.info_high("Initialization Step 1 Complete.")

        # Fix outlet enthalpy and pressure
        for t in blk.flowsheet().time:
            blk.control_volume.properties_out[t].enth_mol.fix(
                value(blk.control_volume.properties_in[t].enth_mol)
                + value(blk.heat_fireside[t])
                / value(blk.control_volume.properties_in[t].flow_mol)
            )
            blk.control_volume.properties_out[t].pressure.fix(
                value(blk.control_volume.properties_in[t].pressure) - 1.0
            )

        blk.heat_eqn.deactivate()
        blk.pressure_change_eqn.deactivate()

        with idaeslog.solver_log(solve_log, idaeslog.DEBUG) as slc:
            res = opt.solve(blk, tee=slc.tee)
        init_log.info_high("Initialization Step 2 {}.".format(idaeslog.condition(res)))

        # Unfix outlet enthalpy and pressure
        for t in blk.flowsheet().time:
            blk.control_volume.properties_out[t].enth_mol.unfix()
            blk.control_volume.properties_out[t].pressure.unfix()
        blk.heat_eqn.activate()
        blk.pressure_change_eqn.activate()

        with idaeslog.solver_log(solve_log, idaeslog.DEBUG) as slc:
            res = opt.solve(blk, tee=slc.tee)
        init_log.info_high("Initialization Step 3 {}.".format(idaeslog.condition(res)))

        blk.control_volume.release_state(flags, outlvl=outlvl)
        init_log.info("Initialization Complete.")

    def calculate_scaling_factors(self):
        super().calculate_scaling_factors()
        for v in self.N_Re.values():
            if iscale.get_scaling_factor(v, warning=True) is None:
                iscale.set_scaling_factor(v, 1e-6)
        for t, c in self.Reynolds_number_eqn.items():
            s = iscale.get_scaling_factor(self.N_Re[t], default=1, warning=True)
            iscale.constraint_scaling_transform(c, s * 1e5, overwrite=False)
        for t, c in self.heat_flux_conv_eqn.items():
            s = iscale.get_scaling_factor(
                self.heat_flux_conv[t], default=1, warning=True
            )
            iscale.constraint_scaling_transform(c, s, overwrite=False)
        for t, c in self.hconv_eqn.items():
            s = iscale.get_scaling_factor(self.hconv[t], default=1, warning=True)
            s *= iscale.get_scaling_factor(self.diameter_in, default=1, warning=True)
            iscale.constraint_scaling_transform(c, s, overwrite=False)
        for t, c in self.pressure_change_eqn.items():
            s = iscale.get_scaling_factor(self.deltaP[t], default=1, warning=True)
            s *= iscale.get_scaling_factor(self.diameter_in, default=1, warning=True)
            iscale.constraint_scaling_transform(c, s, overwrite=False)



#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES).
#
# Copyright (c) 2018-2024 by the software owners: The Regents of the
# University of California, through Lawrence Berkeley National Laboratory,
# National Technology & Engineering Solutions of Sandia, LLC, Carnegie Mellon
# University, West Virginia University Research Corporation, et al.
# All rights reserved.  Please see the files COPYRIGHT.md and LICENSE.md
# for full copyright and license information.
#################################################################################
"""
NTU Condenser model, for use with Helmholtz EOS property packages.
"""
# TODO: Missing docstrings
# pylint: disable=missing-function-docstring

__author__ = "Jinliang Ma"

# Import Pyomo libraries
import pyomo.environ as pyo
from pyomo.environ import units as pyunits
from pyomo.common.config import ConfigBlock, ConfigValue

# Import IDAES cores
from idaes.core import declare_process_block_class, UnitModelBlockData
from idaes.core.util import from_json, to_json, StoreSpec
from idaes.core.solvers import get_solver
from idaes.core.util.tables import create_stream_table_dataframe
from idaes.models.unit_models.heater import (
    _make_heater_config_block,
    _make_heater_control_volume,
)
from idaes.models.unit_models.heat_exchanger import hx_process_config, add_hx_references
import idaes.core.util.scaling as iscale
import idaes.logger as idaeslog


def _make_heat_exchanger_config(config):
    """
    Declare configuration options for HeatExchangerData block.
    """
    config.declare(
        "hot_side_name",
        ConfigValue(
            default="shell",
            domain=str,
            doc="Hot side name, sets control volume and inlet and outlet names",
        ),
    )
    config.declare(
        "cold_side_name",
        ConfigValue(
            default="tube",
            domain=str,
            doc="Cold side name, sets control volume and inlet and outlet names",
        ),
    )
    config.declare(
        "hot_side",
        ConfigBlock(
            implicit=True,
            description="Config block for hot side",
            doc="""A config block used to construct the hot side control volume.
This config can be given by the hot side name instead of hot_side.""",
        ),
    )
    config.declare(
        "cold_side",
        ConfigBlock(
            implicit=True,
            description="Config block for cold side",
            doc="""A config block used to construct the cold side control volume.
This config can be given by the cold side name instead of cold_side.""",
        ),
    )
    _make_heater_config_block(config.hot_side)
    _make_heater_config_block(config.cold_side)


@declare_process_block_class("HelmNtuCondenser", doc="Simple 0D condenser model.")
class HelmNtuCondenserData(UnitModelBlockData):
    """
    Simple NTU condenser unit model.  This model assumes the property packages
    specified are Helmholtz EOS type.
    """

    CONFIG = UnitModelBlockData.CONFIG(implicit=True)
    _make_heat_exchanger_config(CONFIG)

    def build(self):
        """
        Building model

        Args:
            None
        Returns:
            None
        """
        ########################################################################
        #  Call UnitModel.build to setup dynamics and configure                #
        ########################################################################
        super().build()
        hx_process_config(self)
        config = self.config
        time = self.flowsheet().time

        ########################################################################
        # Add control volumes                                                  #
        ########################################################################
        hot_side = _make_heater_control_volume(
            self,
            "hot_side",
            config.hot_side,
            dynamic=config.dynamic,
            has_holdup=config.has_holdup,
        )
        cold_side = _make_heater_control_volume(
            self,
            "cold_side",
            config.cold_side,
            dynamic=config.dynamic,
            has_holdup=config.has_holdup,
        )
        ########################################################################
        # Add variables                                                        #
        ########################################################################
        # Use hot side units as basis
        s1_metadata = self.hot_side.config.property_package.get_metadata()

        f_units = s1_metadata.get_derived_units("flow_mole")
        cp_units = s1_metadata.get_derived_units("heat_capacity_mole")
        q_units = s1_metadata.get_derived_units("power")
        u_units = s1_metadata.get_derived_units("heat_transfer_coefficient")
        a_units = s1_metadata.get_derived_units("area")
        temp_units = s1_metadata.get_derived_units("temperature")

        self.overall_heat_transfer_coefficient = pyo.Var(
            time,
            domain=pyo.PositiveReals,
            initialize=100.0,
            doc="Overall heat transfer coefficient",
            units=u_units,
        )
        self.area = pyo.Var(
            domain=pyo.PositiveReals,
            initialize=1000.0,
            doc="Heat exchange area",
            units=a_units,
        )
        self.heat_duty = pyo.Reference(cold_side.heat)
        ########################################################################
        # Add ports                                                            #
        ########################################################################
        self.add_inlet_port(name="hot_side_inlet", block=hot_side, doc="Hot side inlet")
        self.add_inlet_port(
            name="cold_side_inlet",
            block=cold_side,
            doc="Cold side inlet",
        )
        self.add_outlet_port(
            name="hot_side_outlet", block=hot_side, doc="Hot side outlet"
        )
        self.add_outlet_port(
            name="cold_side_outlet",
            block=cold_side,
            doc="Cold side outlet",
        )
        ########################################################################
        # Add aliases                                                          #
        ########################################################################
        add_hx_references(self)

        ########################################################################
        # Add a unit level energy balance                                      #
        ########################################################################
        @self.Constraint(time, doc="Heat balance equation")
        def unit_heat_balance(b, t):
            return 0 == (
                hot_side.heat[t] + pyunits.convert(cold_side.heat[t], to_units=q_units)
            )

        ########################################################################
        # Add some useful expressions for condenser performance                #
        ########################################################################

        @self.Expression(time, doc="Inlet temperature difference")
        def delta_temperature_in(b, t):
            return hot_side.properties_in[t].temperature - pyunits.convert(
                cold_side.properties_in[t].temperature, temp_units
            )

        @self.Expression(time, doc="Outlet temperature difference")
        def delta_temperature_out(b, t):
            return hot_side.properties_out[t].temperature - pyunits.convert(
                cold_side.properties_out[t].temperature, temp_units
            )

        @self.Expression(time, doc="NTU Based temperature difference")
        def delta_temperature_ntu(b, t):
            return hot_side.properties_in[t].temperature_sat - pyunits.convert(
                cold_side.properties_in[t].temperature, temp_units
            )

        @self.Expression(
            time,
            doc="Minimum product of flow rate and heat "
            "capacity (always on tube side since shell side has phase change)",
        )
        def mcp_min(b, t):
            return pyunits.convert(
                cold_side.properties_in[t].flow_mol
                * cold_side.properties_in[t].cp_mol_phase["Liq"],
                f_units * cp_units,
            )

        @self.Expression(time, doc="Number of transfer units (NTU)")
        def ntu(b, t):
            return b.overall_heat_transfer_coefficient[t] * b.area / b.mcp_min[t]

        @self.Expression(time, doc="Condenser effectiveness factor")
        def effectiveness(b, t):
            return 1 - pyo.exp(-self.ntu[t])

        @self.Expression(time, doc="Heat treansfer")
        def heat_transfer(b, t):
            return b.effectiveness[t] * b.mcp_min[t] * b.delta_temperature_ntu[t]

        ########################################################################
        # Add Equations to calculate heat duty based on NTU method             #
        ########################################################################
        @self.Constraint(time, doc="Heat transfer rate equation based on NTU method")
        def heat_transfer_equation(b, t):
            return pyunits.convert(cold_side.heat[t], q_units) == self.heat_transfer[t]

        @self.Constraint(
            time, doc="Shell side outlet enthalpy is saturated water enthalpy"
        )
        def saturation_eqn(b, t):
            return (
                hot_side.properties_out[t].enth_mol
                == hot_side.properties_in[t].enth_mol_sat_phase["Liq"]
            )

    def set_initial_condition(self):
        if self.config.dynamic is True:
            self.hot_side.material_accumulation[:, :, :].value = 0
            self.hot_side.energy_accumulation[:, :].value = 0
            self.hot_side.material_accumulation[0, :, :].fix(0)
            self.hot_side.energy_accumulation[0, :].fix(0)
            self.cold_side.material_accumulation[:, :, :].value = 0
            self.cold_side.energy_accumulation[:, :].value = 0
            self.cold_side.material_accumulation[0, :, :].fix(0)
            self.cold_side.energy_accumulation[0, :].fix(0)

    def initialize_build(
        self,
        state_args_1=None,
        state_args_2=None,
        unfix="hot_flow",
        outlvl=idaeslog.NOTSET,
        solver=None,
        optarg=None,
    ):
        """
        Condenser initialization method. The initialization routine assumes
        fixed area and heat transfer coefficient and adjusts the cooling water
        flow to condense steam to saturated water at shell side pressure.

        Args:
            state_args_1 : a dict of arguments to be passed to the property
                initialization for hot side (see documentation of the specific
                property package) (default = None).
            state_args_2 : a dict of arguments to be passed to the property
                initialization for cold side (see documentation of the specific
                property package) (default = None).
            optarg : solver options dictionary object (default=None, use
                     default solver options)
            solver : str indicating which solver to use during
                     initialization (default = None, use default solver)

        Returns:
            None

        """
        if unfix not in {"hot_flow", "cold_flow", "pressure"}:
            # PYLINT-TODO
            # pylint: disable-next=broad-exception-raised
            raise Exception(
                "Condenser free variable must be in 'hot_flow', "
                "'cold_flow', or 'pressure'"
            )
        # Set solver options
        init_log = idaeslog.getInitLogger(self.name, outlvl, tag="unit")
        solve_log = idaeslog.getSolveLogger(self.name, outlvl, tag="unit")

        hot_side = getattr(self, self.config.hot_side_name)
        cold_side = getattr(self, self.config.cold_side_name)

        # Store initial model specs, restored at the end of initialization, so
        # the problem is not altered.  This can restore fixed/free vars,
        # active/inactive constraints, and fixed variable values.
        sp = StoreSpec.value_isfixed_isactive(only_fixed=True)
        istate = to_json(self, return_dict=True, wts=sp)

        # Create solver
        opt = get_solver(solver, optarg)

        flags1 = hot_side.initialize(
            outlvl=outlvl, optarg=optarg, solver=solver, state_args=state_args_1
        )
        flags2 = cold_side.initialize(
            outlvl=outlvl, optarg=optarg, solver=solver, state_args=state_args_2
        )
        init_log.info_high("Initialization Step 1 Complete.")

        # Solve with all constraints activated
        self.saturation_eqn.activate()
        if unfix == "pressure":
            hot_side.properties_in[:].pressure.unfix()
        elif unfix == "hot_flow":
            hot_side.properties_in[:].flow_mol.unfix()
        elif unfix == "cold_flow":
            cold_side.properties_in[:].flow_mol.unfix()

        with idaeslog.solver_log(solve_log, idaeslog.DEBUG) as slc:
            res = opt.solve(self, tee=slc.tee)
        init_log.info_high("Initialization Step 4 {}.".format(idaeslog.condition(res)))

        # Release Inlet state
        hot_side.release_state(flags1, outlvl)
        cold_side.release_state(flags2, outlvl)
        from_json(self, sd=istate, wts=sp)

    def _get_performance_contents(self, time_point=0):
        var_dict = {
            "HX Coefficient": self.overall_heat_transfer_coefficient[time_point]
        }
        var_dict["HX Area"] = self.area
        var_dict["Heat Duty"] = self.heat_duty[time_point]
        expr_dict = {}
        expr_dict["Delta T In"] = self.delta_temperature_in[time_point]
        expr_dict["Delta T Out"] = self.delta_temperature_out[time_point]

        return {"vars": var_dict, "exprs": expr_dict}

    def _get_stream_table_contents(self, time_point=0):
        return create_stream_table_dataframe(
            {
                "Hot Inlet": self.hot_side_inlet,
                "Hot Outlet": self.hot_side_outlet,
                "Cold Inlet": self.cold_side_inlet,
                "Cold Outlet": self.cold_side_outlet,
            },
            time_point=time_point,
        )

    def calculate_scaling_factors(self):
        super().calculate_scaling_factors()

        area_sf_default = 1e-2
        overall_heat_transfer_coefficient_sf_default = 1e-2

        # Function to set defaults so I don't need to reproduce the same code
        def _fill_miss_with_default(name, s):
            try:
                c = getattr(self, name)
            except AttributeError:
                return  # it's okay if the attribute doesn't exist, spell careful
            if iscale.get_scaling_factor(c) is None:
                for ci in c.values():
                    if iscale.get_scaling_factor(ci) is None:
                        iscale.set_scaling_factor(ci, s)

        # Set defaults where scale factors are missing
        _fill_miss_with_default("area", area_sf_default)
        _fill_miss_with_default(
            "overall_heat_transfer_coefficient",
            overall_heat_transfer_coefficient_sf_default,
        )

        for t, c in self.heat_transfer_equation.items():
            sf = iscale.get_scaling_factor(self.cold_side.heat[t])
            iscale.constraint_scaling_transform(c, sf, overwrite=False)

        for t, c in self.unit_heat_balance.items():
            sf = iscale.get_scaling_factor(self.cold_side.heat[t])
            iscale.constraint_scaling_transform(c, sf, overwrite=False)

        for t, c in self.saturation_eqn.items():
            sf = iscale.get_scaling_factor(self.hot_side.properties_out[t].enth_mol)
            iscale.constraint_scaling_transform(c, sf, overwrite=False)



#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES).
#
# Copyright (c) 2018-2024 by the software owners: The Regents of the
# University of California, through Lawrence Berkeley National Laboratory,
# National Technology & Engineering Solutions of Sandia, LLC, Carnegie Mellon
# University, West Virginia University Research Corporation, et al.
# All rights reserved.  Please see the files COPYRIGHT.md and LICENSE.md
# for full copyright and license information.
#################################################################################
"""
Pure component properties as used by the NIST WebBook

https://webbook.nist.gov/chemistry/

Retrieved: September 13th, 2019

All parameter indices and units based on conventions used by the source
"""
# TODO: Missing doc strings
# pylint: disable=missing-class-docstring
# pylint: disable=missing-function-docstring

from pyomo.environ import Expression, log, Var, units as pyunits

from idaes.core.util.misc import set_param_from_config


# -----------------------------------------------------------------------------
class NIST(object):
    # -----------------------------------------------------------------------------
    # Shomate Equation for heat capacities, enthalpy and entropy
    class cp_mol_ig_comp:
        @staticmethod
        def build_parameters(cobj):
            cobj.cp_mol_ig_comp_coeff_A = Var(
                doc="Shomate A parameter for ideal gas molar heat capacity",
                units=pyunits.J * pyunits.mol**-1 * pyunits.K**-1,
            )
            set_param_from_config(cobj, param="cp_mol_ig_comp_coeff", index="A")

            cobj.cp_mol_ig_comp_coeff_B = Var(
                doc="Shomate B parameter for ideal gas molar heat capacity",
                units=pyunits.J * pyunits.mol**-1 * pyunits.K**-1 * pyunits.kiloK**-1,
            )
            set_param_from_config(cobj, param="cp_mol_ig_comp_coeff", index="B")

            cobj.cp_mol_ig_comp_coeff_C = Var(
                doc="Shomate C parameter for ideal gas molar heat capacity",
                units=pyunits.J * pyunits.mol**-1 * pyunits.K**-1 * pyunits.kiloK**-2,
            )
            set_param_from_config(cobj, param="cp_mol_ig_comp_coeff", index="C")

            cobj.cp_mol_ig_comp_coeff_D = Var(
                doc="Shomate D parameter for ideal gas molar heat capacity",
                units=pyunits.J * pyunits.mol**-1 * pyunits.K**-1 * pyunits.kiloK**-3,
            )
            set_param_from_config(cobj, param="cp_mol_ig_comp_coeff", index="D")

            cobj.cp_mol_ig_comp_coeff_E = Var(
                doc="Shomate E parameter for ideal gas molar heat capacity",
                units=pyunits.J * pyunits.mol**-1 * pyunits.K**-1 * pyunits.kiloK**2,
            )
            set_param_from_config(cobj, param="cp_mol_ig_comp_coeff", index="E")

            cobj.cp_mol_ig_comp_coeff_F = Var(
                doc="Shomate F parameter for ideal gas molar heat capacity",
                units=pyunits.kJ * pyunits.mol**-1,
            )
            set_param_from_config(cobj, param="cp_mol_ig_comp_coeff", index="F")

            cobj.cp_mol_ig_comp_coeff_G = Var(
                doc="Shomate G parameter for ideal gas molar heat capacity",
                units=pyunits.J * pyunits.mol**-1 * pyunits.K**-1,
            )
            set_param_from_config(cobj, param="cp_mol_ig_comp_coeff", index="G")

            cobj.cp_mol_ig_comp_coeff_H = Var(
                doc="Shomate H parameter for ideal gas molar heat capacity",
                units=pyunits.kJ * pyunits.mol**-1,
            )
            set_param_from_config(cobj, param="cp_mol_ig_comp_coeff", index="H")

            # As the H parameter is the specific heat of formation, create an
            # Expression which converts this to the base units with standard name
            units = cobj.parent_block().get_metadata().derived_units
            cobj.enth_mol_form_vap_comp_ref = Expression(
                expr=pyunits.convert(
                    cobj.cp_mol_ig_comp_coeff_H, to_units=units.ENERGY_MOLE
                ),
                doc="Vapor phase molar heat of formation @ Tref",
            )

        @staticmethod
        def return_expression(b, cobj, T):
            # Specific heat capacity (const. P)  via the Shomate equation
            t = pyunits.convert(T, to_units=pyunits.kiloK)
            cp = (
                cobj.cp_mol_ig_comp_coeff_A
                + cobj.cp_mol_ig_comp_coeff_B * t
                + cobj.cp_mol_ig_comp_coeff_C * t**2
                + cobj.cp_mol_ig_comp_coeff_D * t**3
                + cobj.cp_mol_ig_comp_coeff_E * t**-2
            )

            units = b.params.get_metadata().derived_units
            return pyunits.convert(cp, units.HEAT_CAPACITY_MOLE)

    class enth_mol_ig_comp:
        @staticmethod
        def build_parameters(cobj):
            if not hasattr(cobj, "cp_mol_ig_comp_coeff_A"):
                NIST.cp_mol_ig_comp.build_parameters(cobj)

        @staticmethod
        def return_expression(b, cobj, T):
            # Specific enthalpy via the Shomate equation
            t = pyunits.convert(T, to_units=pyunits.kiloK)

            h_form = b.params.config.include_enthalpy_of_formation
            H = (
                cobj.cp_mol_ig_comp_coeff_H
                if not h_form
                else 0 * pyunits.kJ * pyunits.mol**-1
            )

            h = (
                cobj.cp_mol_ig_comp_coeff_A * t
                + (cobj.cp_mol_ig_comp_coeff_B / 2) * t**2
                + (cobj.cp_mol_ig_comp_coeff_C / 3) * t**3
                + (cobj.cp_mol_ig_comp_coeff_D / 4) * t**4
                - cobj.cp_mol_ig_comp_coeff_E / t
                + cobj.cp_mol_ig_comp_coeff_F
                - H
            )

            units = b.params.get_metadata().derived_units
            return pyunits.convert(h, units.ENERGY_MOLE)

    class entr_mol_ig_comp:
        @staticmethod
        def build_parameters(cobj):
            if not hasattr(cobj, "cp_mol_ig_comp_coeff_A"):
                NIST.cp_mol_ig_comp.build_parameters(cobj)

        @staticmethod
        def return_expression(b, cobj, T):
            # Specific entropy via the Shomate equation
            t = pyunits.convert(T, to_units=pyunits.kiloK)

            s = (
                cobj.cp_mol_ig_comp_coeff_A * log(t / pyunits.kiloK)
                + cobj.cp_mol_ig_comp_coeff_B * t  # need to make unitless
                + (cobj.cp_mol_ig_comp_coeff_C / 2) * t**2
                + (cobj.cp_mol_ig_comp_coeff_D / 3) * t**3
                - (cobj.cp_mol_ig_comp_coeff_E / 2) * t**-2
                + cobj.cp_mol_ig_comp_coeff_G
            )

            units = b.params.get_metadata().derived_units
            return pyunits.convert(s, units.ENTROPY_MOLE)

    # -----------------------------------------------------------------------------
    # Antoine equation for saturation pressure
    class pressure_sat_comp:
        @staticmethod
        def build_parameters(cobj):
            cobj.pressure_sat_comp_coeff_A = Var(
                doc="Antoine A coefficient for calculating Psat", units=None
            )
            set_param_from_config(cobj, param="pressure_sat_comp_coeff", index="A")

            cobj.pressure_sat_comp_coeff_B = Var(
                doc="Antoine B coefficient for calculating Psat", units=pyunits.K
            )
            set_param_from_config(cobj, param="pressure_sat_comp_coeff", index="B")

            cobj.pressure_sat_comp_coeff_C = Var(
                doc="Antoine C coefficient for calculating Psat", units=pyunits.K
            )
            set_param_from_config(cobj, param="pressure_sat_comp_coeff", index="C")

        @staticmethod
        def return_expression(b, cobj, T, dT=False):
            if dT:
                return NIST.pressure_sat_comp.dT_expression(b, cobj, T)

            psat = (
                10
                ** (
                    cobj.pressure_sat_comp_coeff_A
                    - cobj.pressure_sat_comp_coeff_B
                    / (
                        pyunits.convert(T, to_units=pyunits.K)
                        + cobj.pressure_sat_comp_coeff_C
                    )
                )
                * pyunits.bar
            )

            units = b.params.get_metadata().derived_units
            return pyunits.convert(psat, to_units=units.PRESSURE)

        @staticmethod
        def dT_expression(b, cobj, T):
            p_sat_dT = (
                NIST.pressure_sat_comp.return_expression(b, cobj, T)
                * cobj.pressure_sat_comp_coeff_B
                * log(10)
                / (
                    pyunits.convert(T, to_units=pyunits.K)
                    + cobj.pressure_sat_comp_coeff_C
                )
                ** 2
            )

            units = b.params.get_metadata().derived_units
            dp_units = units.PRESSURE / units.TEMPERATURE
            return pyunits.convert(p_sat_dT, to_units=dp_units)



#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES).
#
# Copyright (c) 2018-2024 by the software owners: The Regents of the
# University of California, through Lawrence Berkeley National Laboratory,
# National Technology & Engineering Solutions of Sandia, LLC, Carnegie Mellon
# University, West Virginia University Research Corporation, et al.
# All rights reserved.  Please see the files COPYRIGHT.md and LICENSE.md
# for full copyright and license information.
#################################################################################
"""
Model for channel in SolidOxideCell. If opposite_flow is false, material flows from
``z=0`` to ``z=1``, otherwise it flows from ``z=1`` to ``z=0``. Pure convection occurs in the z
direction, while mass and heat transfer coefficients govern fluxes out of the ``x=0``
and ``x=1`` sides of the model. The PDEs are discretized in the z direction  by a finite
volume method with upwind differences used to calculate convection fluxes. It is
assumed that the gas is ideal.

Like most submodels, the SocChannel will create terms for temperature and heat flux
at both x=0 and x=1 edges if they are not provided in the config. However,
concentration variables and xflux terms are created only on the side facing the
electrode (``x=1`` if below_electrode is true, ``x=0`` if false). Expressions are created
for the other side, to avoid additional logic with material balances.

Manipulated variables (typically fixed through cell-level ports):
    - ``flow_mol_inlet[t]``: Molar flow into channel
    - ``mole_frac_comp_inlet[t, j]``: Molar composition of flow into channel
    - ``temperature_inlet[t]``: Channel inlet flow rate
    - ``pressure_inlet[t]``: Channel inlet pressure

Boundary variables:
    - ``temperature_deviation_x0[t, iz]``
    - ``temperature_deviation_x1[t, iz]``
    - ``heat_flux_x0[t, iz]``
    - ``heat_flux_x1[t, iz]``

    * If ``below_electrode = False``:
        - ``material_flux_x0[t, iz]``
        - ``conc_mol_comp_deviation_x0[t, iz]``

    * If ``below_electrode = True``:
        - ``material_flux_x1[t, iz]``
        - ``conc_mol_comp_deviation_x1[t, iz]``

Instances of ``Var`` that must be fixed:
    - ``length_x``: Distance between electrode and interconnect
    - ``heat_transfer_coefficient[t, iz]``: Heat transfer coefficient between channel bulk and surfaces.
      Correlations to calculate this are planned, but they will add a large number of equations.
"""
# TODO: Missing docstrings
# pylint: disable=missing-class-docstring
# pylint: disable=missing-function-docstring

# TODO: Look into protected access issues
# pylint: disable=protected-access

__author__ = "John Eslick, Douglas Allan"

from pyomo.common.config import ConfigValue, In, ListOf, Bool
from pyomo.dae import DerivativeVar
import pyomo.environ as pyo

from idaes.core import declare_process_block_class, UnitModelBlockData, useDefault
from idaes.core.util.constants import Constants
import idaes.models_extra.power_generation.unit_models.soc_submodels.common as common
from idaes.models_extra.power_generation.unit_models.soc_submodels.common import (
    _set_if_unfixed,
    _gas_species_list,
    _element_list,
    _element_dict,
)
import idaes.core.util.scaling as iscale
from idaes.core.solvers import get_solver

from idaes.core.base.var_like_expression import VarLikeExpression
import idaes.logger as idaeslog


@declare_process_block_class("SocChannel")
class SocChannelData(UnitModelBlockData):
    CONFIG = UnitModelBlockData.CONFIG()
    CONFIG.declare(
        "component_list",
        ConfigValue(
            domain=common._SubsetOf(_gas_species_list), description="List of components"
        ),
    )

    CONFIG.declare(
        "control_volume_zfaces",
        ConfigValue(
            domain=ListOf(float),
            description="List containing coordinates of control volume faces "
            "in z direction. Coordinates must start with zero, be strictly "
            "increasing, and end with one",
        ),
    )
    CONFIG.declare(
        "length_z",
        ConfigValue(
            default=None, description="Length in the direction of flow (z-direction)"
        ),
    )
    CONFIG.declare(
        "length_y", ConfigValue(default=None, description="Width of cell (y-direction)")
    )
    CONFIG.declare(
        "include_temperature_x_thermo",
        ConfigValue(
            domain=In([useDefault, True, False]),
            default=True,
            description="Whether to consider temperature variations in "
            "x direction in thermodynamic equations",
        ),
    )
    common._thermal_boundary_conditions_config(CONFIG, thin=False)
    CONFIG.declare(
        "opposite_flow",
        ConfigValue(
            domain=Bool,
            default=False,
            description="If True, gas flows from z=1 to z=0. Otherwise "
            "gas flows from z=0 to z=1.",
        ),
    )
    CONFIG.declare(
        "below_electrode",
        ConfigValue(
            domain=Bool,
            description="If True, the top of the channel is the electrode "
            "boundary and material flux terms are created there. If False "
            "the bottom is the electrode boundary and material flux terms "
            "are created there.",
        ),
    )

    def build(self):
        super().build()

        # Set up some sets for the space and time indexing
        dynamic = self.config.dynamic
        tset = self.flowsheet().config.time
        self.component_list = pyo.Set(
            initialize=self.config.component_list,
            ordered=True,
            doc="Set of all components present in submodel",
        )
        # Set up node and face sets and get integer indices for them
        izfaces, iznodes = common._face_initializer(
            self, self.config.control_volume_zfaces, "z"
        )
        # Space saving aliases
        comps = self.component_list
        include_temp_x_thermo = self.config.include_temperature_x_thermo

        common._create_if_none(self, "length_z", idx_set=None, units=pyo.units.m)
        common._create_if_none(self, "length_y", idx_set=None, units=pyo.units.m)

        common._create_thermal_boundary_conditions_if_none(self, thin=False)

        if not self.config.below_electrode:
            self.conc_mol_comp_deviation_x0 = pyo.Var(
                tset,
                iznodes,
                comps,
                doc="Deviation of concentration at electrode surface from that "
                "in the bulk channel",
                initialize=0,
                units=pyo.units.mol / pyo.units.m**3,
            )
            self.material_flux_x0 = pyo.Var(
                tset,
                iznodes,
                comps,
                doc="Material flux from electrode surface to channel "
                "(positive is in)",
                initialize=0,
                units=pyo.units.mol / pyo.units.m**2 / pyo.units.s,
            )

            self.conc_mol_comp_deviation_x1 = pyo.Param(
                tset,
                iznodes,
                comps,
                doc="Dummy parameter for zero concentration deviation at interconnect wall.",
                initialize=0,
                units=pyo.units.mol / pyo.units.m**3,
            )
            self.material_flux_x1 = pyo.Param(
                tset,
                iznodes,
                comps,
                doc="Dummy parameter for zero material flux from channel "
                "to interconnect.",
                initialize=0,
                units=pyo.units.mol / pyo.units.m**2 / pyo.units.s,
            )

        else:
            self.conc_mol_comp_deviation_x1 = pyo.Var(
                tset,
                iznodes,
                comps,
                doc="Deviation of concentration at electrode surface from that "
                "in the bulk channel",
                initialize=0,
                units=pyo.units.mol / pyo.units.m**3,
            )
            self.material_flux_x1 = pyo.Var(
                tset,
                iznodes,
                comps,
                doc="Material flux from channel to electrode surface "
                "(positive is out)",
                initialize=0,
                units=pyo.units.mol / pyo.units.m**2 / pyo.units.s,
            )
            self.conc_mol_comp_deviation_x0 = pyo.Param(
                tset,
                iznodes,
                comps,
                doc="Dummy parameter for zero concentration deviation at interconnect wall.",
                initialize=0,
                units=pyo.units.mol / pyo.units.m**3,
            )
            self.material_flux_x0 = pyo.Param(
                tset,
                iznodes,
                comps,
                doc="Dummy parameter for zero material flux from channel "
                "to interconnect.",
                initialize=0,
                units=pyo.units.mol / pyo.units.m**2 / pyo.units.s,
            )

        # Channel thickness AKA length in the x direction is specific to the
        # channel so local variable here is the only option
        self.length_x = pyo.Var(
            doc="Thickness from interconnect to electrode (x-direction)",
            units=pyo.units.m,
        )
        self.heat_transfer_coefficient = pyo.Var(
            tset,
            iznodes,
            doc="Local channel heat transfer coefficient",
            initialize=500,
            units=pyo.units.J / pyo.units.m**2 / pyo.units.s / pyo.units.K,
        )
        self.flow_mol = pyo.Var(
            tset,
            iznodes,
            doc="Molar flow in the z-direction through faces",
            units=pyo.units.mol / pyo.units.s,
            bounds=(-1e-9, None),
        )
        self.conc_mol_comp = pyo.Var(
            tset,
            iznodes,
            comps,
            doc="Component concentration at node centers",
            units=pyo.units.mol / pyo.units.m**3,
            bounds=(0, None),
        )
        self.temperature_deviation_x = pyo.Var(
            tset,
            iznodes,
            doc="Deviation of temperature at node centers " "from temperature_z",
            units=pyo.units.K,
            bounds=(-1000, 1000),
        )
        self.enth_mol = pyo.Var(
            tset,
            iznodes,
            doc="Molar enthalpy at node centers",
            units=pyo.units.J / pyo.units.mol,
        )
        if self.config.has_holdup:
            self.int_energy_mol = pyo.Var(
                tset,
                iznodes,
                doc="Molar internal energy at node centers",
                units=pyo.units.J / pyo.units.mol,
            )
            self.int_energy_density = pyo.Var(
                tset,
                iznodes,
                doc="Molar internal energy density at node centers",
                units=pyo.units.J / pyo.units.m**3,
            )
        self.velocity = pyo.Var(
            tset,
            iznodes,
            doc="Fluid velocity at node centers",
            units=pyo.units.m / pyo.units.s,
        )
        self.pressure = pyo.Var(
            tset,
            iznodes,
            doc="Pressure at node centers",
            units=pyo.units.Pa,
            bounds=(0, None),
        )
        self.mole_frac_comp = pyo.Var(
            tset,
            iznodes,
            comps,
            doc="Component mole fraction at node centers",
            bounds=(0, None),
            units=pyo.units.dimensionless,
        )
        self.diff_eff_coeff = pyo.Var(
            tset,
            iznodes,
            comps,
            doc="Effective Fick's law diffusion coefficient at node centers",
            initialize=2e-5,
            bounds=(0, None),
            units=pyo.units.m**2 / pyo.units.s,
        )
        self.flow_mol_inlet = pyo.Var(
            tset,
            doc="Inlet face molar flow rate",
            bounds=(0, None),
            units=pyo.units.mol / pyo.units.s,
        )
        self.pressure_inlet = pyo.Var(
            tset, doc="Inlet pressure", bounds=(0, None), units=pyo.units.Pa
        )
        self.temperature_inlet = pyo.Var(
            tset, doc="Inlet temperature", bounds=(300, None), units=pyo.units.K
        )
        self.temperature_outlet = pyo.Var(
            tset, doc="Outlet temperature", bounds=(300, None), units=pyo.units.K
        )
        self.mole_frac_comp_inlet = pyo.Var(
            tset,
            comps,
            doc="Inlet component mole fractions",
            bounds=(0, 1),
            units=pyo.units.dimensionless,
        )

        # Add time derivative variable if steady state use const 0.
        if dynamic:
            self.dconc_mol_compdt = DerivativeVar(
                self.conc_mol_comp,
                wrt=tset,
                initialize=0,
                doc="Component concentration time derivative",
            )
        else:
            self.dconc_mol_compdt = pyo.Param(
                tset,
                iznodes,
                comps,
                initialize=0,
                units=pyo.units.mol / pyo.units.m**3 / pyo.units.s,
            )

        # Add time derivative variable if steady state use const 0.
        if dynamic:
            self.dcedt = DerivativeVar(
                self.int_energy_density,
                wrt=tset,
                initialize=0,
                doc="Internal energy density time derivative",
            )
        else:
            self.dcedt = pyo.Param(
                tset,
                iznodes,
                initialize=0,
                units=pyo.units.J / pyo.units.m**3 / pyo.units.s,
            )

        @self.Expression()
        def flow_area(b):
            return b.length_x[None] * b.length_y[None]

        @self.Expression(iznodes)
        def dz(b, iz):
            return b.zfaces.at(iz + 1) - b.zfaces.at(iz)

        @self.Expression(iznodes)
        def node_volume(b, iz):
            return b.length_x[None] * b.length_y[None] * b.length_z[None] * b.dz[iz]

        @self.Expression(iznodes)
        def xface_area(b, iz):
            return b.length_z[None] * b.length_y[None] * b.dz[iz]

        @self.Expression(tset, iznodes)
        def temperature(b, t, iz):
            if include_temp_x_thermo:
                return b.temperature_z[t, iz] + b.temperature_deviation_x[t, iz]
            else:
                return b.temperature_z[t, iz]

        @self.Expression(tset, iznodes)
        def vol_mol(b, t, iz):
            return Constants.gas_constant * b.temperature[t, iz] / b.pressure[t, iz]

        @self.Expression(tset)
        def vol_mol_inlet(b, t):
            return Constants.gas_constant * b.temperature_inlet[t] / b.pressure_inlet[t]

        # TODO maybe replace with variable-constraint pair?
        @self.Expression(tset)
        def enth_mol_inlet(b, t):
            return sum(
                common._comp_enthalpy_expr(b.temperature_inlet[t], i)
                * b.mole_frac_comp_inlet[t, i]
                for i in comps
            )

        @self.Constraint(tset, iznodes, comps)
        def diff_eff_coeff_eqn(b, t, iz, i):
            T = b.temperature[t, iz]
            P = b.pressure[t, iz]
            x = b.mole_frac_comp
            bfun = common._binary_diffusion_coefficient_expr
            return b.diff_eff_coeff[t, iz, i] == (1.0 - x[t, iz, i]) / sum(
                x[t, iz, j] / bfun(T, P, i, j) for j in comps if i != j
            )

        @self.Expression(tset, iznodes, comps)
        def mass_transfer_coeff(b, t, iz, i):
            # Quick and dirty approximation based on Ficks law through a thin
            # film of length L_x/2. For small concentration gradients
            # this will (hopefully) be enough
            return 2 * b.diff_eff_coeff[t, iz, i] / b.length_x

        if not self.config.below_electrode:

            @self.Constraint(tset, iznodes, comps)
            def material_flux_x0_eqn(b, t, iz, i):
                return (
                    b.material_flux_x0[t, iz, i] / b.mass_transfer_coeff[t, iz, i]
                    == b.conc_mol_comp_deviation_x0[t, iz, i]
                )

        else:

            @self.Constraint(tset, iznodes, comps)
            def material_flux_x1_eqn(b, t, iz, i):
                return (
                    -b.material_flux_x1[t, iz, i] / b.mass_transfer_coeff[t, iz, i]
                    == b.conc_mol_comp_deviation_x1[t, iz, i]
                )

        @self.Constraint(tset, iznodes)
        def flow_mol_eqn(b, t, iz):
            # either way the flow goes, want the flow rate to be positive, but
            # in the opposite flow cases want flux and velocity to be negative
            if self.config.opposite_flow:
                return (
                    b.flow_mol[t, iz]
                    == -b.flow_area * b.velocity[t, iz] / b.vol_mol[t, iz]
                )
            return (
                b.flow_mol[t, iz] == b.flow_area * b.velocity[t, iz] / b.vol_mol[t, iz]
            )

        @self.Constraint(tset, iznodes)
        def constant_pressure_eqn(b, t, iz):
            return b.pressure[t, iz] == b.pressure_inlet[t]

        @self.Constraint(tset, iznodes, comps)
        def conc_mol_comp_eqn(b, t, iz, i):
            return (
                b.conc_mol_comp[t, iz, i]
                * b.temperature[t, iz]
                * Constants.gas_constant
                == b.pressure[t, iz] * b.mole_frac_comp[t, iz, i]
            )

        @self.Constraint(tset, iznodes)
        def enth_mol_eqn(b, t, iz):
            return b.enth_mol[t, iz] == sum(
                common._comp_enthalpy_expr(b.temperature[t, iz], i)
                * b.mole_frac_comp[t, iz, i]
                for i in comps
            )

        if self.config.has_holdup:

            @self.Constraint(tset, iznodes)
            def int_energy_mol_eqn(b, t, iz):
                return b.int_energy_mol[t, iz] == sum(
                    common._comp_int_energy_expr(b.temperature[t, iz], i)
                    * b.mole_frac_comp[t, iz, i]
                    for i in comps
                )

            @self.Constraint(tset, iznodes)
            def int_energy_density_eqn(b, t, iz):
                return (
                    b.int_energy_density[t, iz]
                    == b.int_energy_mol[t, iz] / b.vol_mol[t, iz]
                )

        @self.Constraint(tset, iznodes)
        def mole_frac_comp_eqn(b, t, iz):
            return 1 == sum(b.mole_frac_comp[t, iz, i] for i in comps)

        @self.Expression(tset, comps)
        def flow_mol_comp_inlet(b, t, i):
            return b.flow_mol_inlet[t] * b.mole_frac_comp_inlet[t, i]

        @self.Expression(tset, comps)
        def material_flux_z_inlet(b, t, i):
            # either way the flow goes, want the flow rate to be positive, but
            # in the opposite flow cases want flux and velocity to be negative
            if self.config.opposite_flow:
                return -b.flow_mol_inlet[t] / b.flow_area * b.mole_frac_comp_inlet[t, i]
            return b.flow_mol_inlet[t] / b.flow_area * b.mole_frac_comp_inlet[t, i]

        @self.Expression(tset)
        def enth_flux_z_inlet(b, t):
            # either way the flow goes, want the flow rate to be positive, but
            # in the opposite flow cases want flux and velocity to be negative
            if self.config.opposite_flow:
                return -b.flow_mol_inlet[t] / b.flow_area * b.enth_mol_inlet[t]
            return b.flow_mol_inlet[t] / b.flow_area * b.enth_mol_inlet[t]

        @self.Expression(tset, izfaces, comps)
        def material_flux_z(b, t, iz, i):
            return common._interpolate_channel(
                iz=iz,
                ifaces=izfaces,
                nodes=b.znodes,
                faces=b.zfaces,
                phi_func=lambda iface: b.velocity[t, iface]
                * b.conc_mol_comp[t, iface, i],
                phi_inlet=b.material_flux_z_inlet[t, i],
                opposite_flow=self.config.opposite_flow,
            )

        @self.Expression(tset, izfaces)
        def material_flux_z_enth(b, t, iz):
            return common._interpolate_channel(
                iz=iz,
                ifaces=izfaces,
                nodes=b.znodes,
                faces=b.zfaces,
                phi_func=lambda iface: b.velocity[t, iface]
                / b.vol_mol[t, iface]
                * b.enth_mol[t, iface],
                phi_inlet=b.enth_flux_z_inlet[t],
                opposite_flow=self.config.opposite_flow,
            )

        @self.Expression(tset, izfaces)
        def pressure_face(b, t, iz):
            # Although I'm currently assuming no pressure drop in the channel
            # and don't have a momentum balance, this will let me estimate the
            # outlet pressure in a way that will let me add in a momentum balance
            # later
            return common._interpolate_channel(
                iz=iz,
                ifaces=izfaces,
                nodes=b.znodes,
                faces=b.zfaces,
                phi_func=lambda iface: b.pressure[t, iface],
                phi_inlet=b.pressure_inlet[t],
                opposite_flow=self.config.opposite_flow,
            )

        @self.Constraint(tset, iznodes, comps)
        def material_balance_eqn(b, t, iz, i):
            return b.dconc_mol_compdt[t, iz, i] * b.node_volume[iz] == b.flow_area * (
                b.material_flux_z[t, iz, i] - b.material_flux_z[t, iz + 1, i]
            ) + b.xface_area[iz] * (
                b.material_flux_x0[t, iz, i] - b.material_flux_x1[t, iz, i]
            )

        if dynamic:
            self.material_balance_eqn[tset.first(), :, :].deactivate()

        @self.Constraint(tset, iznodes)
        def energy_balance_eqn(b, t, iz):
            return (
                b.dcedt[t, iz] * b.node_volume[iz]
                == b.flow_area
                * (b.material_flux_z_enth[t, iz] - b.material_flux_z_enth[t, iz + 1])
                + b.xface_area[iz]
                * sum(
                    b.material_flux_x0[t, iz, i]
                    * common._comp_enthalpy_expr(b.temperature_x0[t, iz], i)
                    - b.material_flux_x1[t, iz, i]
                    * common._comp_enthalpy_expr(b.temperature_x1[t, iz], i)
                    for i in comps
                )
                + b.heat_flux_x0[t, iz] * b.xface_area[iz]
                - b.heat_flux_x1[t, iz] * b.xface_area[iz]
            )

        if dynamic:
            self.energy_balance_eqn[tset.first(), :].deactivate()

        @self.Constraint(tset, iznodes)
        def temperature_x0_eqn(b, t, iz):
            return self.heat_flux_x0[t, iz] == self.heat_transfer_coefficient[t, iz] * (
                self.temperature_deviation_x0[t, iz]
                - self.temperature_deviation_x[t, iz]
            )

        @self.Constraint(tset, iznodes)
        def temperature_x1_eqn(b, t, iz):
            return self.heat_flux_x1[t, iz] == self.heat_transfer_coefficient[t, iz] * (
                self.temperature_deviation_x[t, iz]
                - self.temperature_deviation_x1[t, iz]
            )

        # For convenience define outlet expressions
        if self.config.opposite_flow:
            izfout = self.izfout = izfaces.first()
        else:
            izfout = self.izfout = izfaces.last()

        @self.Expression(tset, comps)
        def flow_mol_comp_outlet(b, t, i):
            if self.config.opposite_flow:
                return -b.material_flux_z[t, izfout, i] * b.flow_area
            else:
                return b.material_flux_z[t, izfout, i] * b.flow_area

        def rule_flow_mol_outlet(b, t):
            return sum(b.flow_mol_comp_outlet[t, i] for i in comps)

        def rule_pressure_outlet(b, t):
            return b.pressure_face[t, izfout]

        def rule_mole_frac_comp_outlet(b, t, i):
            return b.flow_mol_comp_outlet[t, i] / b.flow_mol_outlet[t]

        self.flow_mol_outlet = VarLikeExpression(tset, rule=rule_flow_mol_outlet)
        self.pressure_outlet = VarLikeExpression(tset, rule=rule_pressure_outlet)
        self.mole_frac_comp_outlet = VarLikeExpression(
            tset, comps, rule=rule_mole_frac_comp_outlet
        )

        @self.Expression(tset)
        def enth_mol_outlet(b, t):
            if self.config.opposite_flow:
                return (
                    -b.material_flux_z_enth[t, izfout]
                    * b.flow_area
                    / b.flow_mol_outlet[t]
                )
            else:
                return (
                    b.material_flux_z_enth[t, izfout]
                    * b.flow_area
                    / b.flow_mol_outlet[t]
                )

        # know enthalpy need a constraint to back calculate temperature
        @self.Constraint(tset)
        def temperature_outlet_eqn(b, t):
            return b.enth_mol_outlet[t] == sum(
                common._comp_enthalpy_expr(b.temperature_outlet[t], i)
                * b.mole_frac_comp_outlet[t, i]
                for i in comps
            )

    def initialize_build(
        self,
        outlvl=idaeslog.NOTSET,
        solver=None,
        optarg=None,
    ):
        # At present, this method does not fix inlet variables because they are
        # fixed at the cell level instead.
        # TODO Add ports to submodel instead?

        solve_log = idaeslog.getSolveLogger(self.name, outlvl, tag="unit")

        tset = self.flowsheet().config.time

        for t in tset:
            _set_if_unfixed(self.temperature_outlet[t], self.temperature_inlet[t])
            for iz in self.iznodes:
                _set_if_unfixed(self.temperature_z[t, iz], self.temperature_inlet[t])
                _set_if_unfixed(self.temperature_deviation_x0[t, iz], 0)
                _set_if_unfixed(self.temperature_deviation_x[t, iz], 0)
                _set_if_unfixed(self.temperature_deviation_x1[t, iz], 0)
                _set_if_unfixed(self.pressure[t, iz], self.pressure_inlet[t])
                _set_if_unfixed(self.flow_mol[t, iz], self.flow_mol_inlet[t])
                for i in self.component_list:
                    _set_if_unfixed(
                        self.mole_frac_comp[t, iz, i], self.mole_frac_comp_inlet[t, i]
                    )
                    _set_if_unfixed(
                        self.conc_mol_comp[t, iz, i],
                        self.pressure[t, iz]
                        * self.mole_frac_comp[t, iz, i]
                        / self.temperature[t, iz]
                        / Constants.gas_constant,
                    )
                _set_if_unfixed(
                    self.velocity[t, iz],
                    self.flow_mol[t, iz] / self.flow_area * self.vol_mol[t, iz],
                )
                _set_if_unfixed(
                    self.enth_mol[t, iz],
                    sum(
                        common._comp_enthalpy_expr(self.temperature[t, iz], i)
                        * self.mole_frac_comp[t, iz, i]
                        for i in self.component_list
                    ),
                )
                if self.config.has_holdup:
                    _set_if_unfixed(
                        self.int_energy_mol[t, iz],
                        sum(
                            common._comp_int_energy_expr(self.temperature[t, iz], i)
                            * self.mole_frac_comp[t, iz, i]
                            for i in self.component_list
                        ),
                    )
                    _set_if_unfixed(
                        self.int_energy_density[t, iz],
                        self.int_energy_mol[t, iz] / self.vol_mol[t, iz],
                    )
        solver_obj = get_solver(solver, optarg)
        common._init_solve_block(self, solver_obj, solve_log)

    def calculate_scaling_factors(self):
        pass

    def model_check(self, steady_state=True):
        if not steady_state:
            # Mass and energy conservation equations steady state only at present
            return

        comp_set = set(self.component_list)
        elements_present = set()

        for element in _element_list:
            include_element = False
            for species in _gas_species_list:
                # Floating point equality take warning!
                if species in comp_set and _element_dict[element][species] != 0:
                    include_element = True
            if include_element:
                elements_present.add(element)

        for t in self.flowsheet().config.time:
            for element in _element_list:
                if element not in elements_present:
                    continue
                sum_in = sum(
                    _element_dict[element][j] * self.flow_mol_comp_inlet[t, j]
                    for j in self.component_list
                )
                sum_out = sum(
                    _element_dict[element][j] * self.flow_mol_comp_outlet[t, j]
                    for j in self.component_list
                )
                for iz in self.iznodes:
                    sum_in += sum(
                        _element_dict[element][j]
                        * self.material_flux_x0[t, iz, j]
                        * self.xface_area[iz]
                        for j in self.component_list
                    )
                    sum_out += sum(
                        _element_dict[element][j]
                        * self.material_flux_x1[t, iz, j]
                        * self.xface_area[iz]
                        for j in self.component_list
                    )
                normal = max(
                    pyo.value(sum_in), pyo.value(sum_out), 1e-9
                )  # FIXME justify this number
                fraction_change = pyo.value((sum_out - sum_in) / normal)
                if abs(fraction_change) > 1e-5:
                    raise RuntimeError(
                        f"{element} is not being conserved in {self.name}; "
                        f"fractional change {fraction_change}."
                    )
            enth_in = self.enth_mol_inlet[t] * self.flow_mol_inlet[t]
            enth_out = self.enth_mol_outlet[t] * self.flow_mol_outlet[t]

            for iz in self.iznodes:
                enth_in += self.xface_area[iz] * (
                    self.heat_flux_x0[t, iz]
                    + sum(
                        common._comp_enthalpy_expr(self.temperature_x0[t, iz], j)
                        * self.material_flux_x0[t, iz, j]
                        for j in self.component_list
                    )
                )
                enth_out += self.xface_area[iz] * (
                    self.heat_flux_x1[t, iz]
                    + sum(
                        common._comp_enthalpy_expr(self.temperature_x1[t, iz], j)
                        * self.material_flux_x1[t, iz, j]
                        for j in self.component_list
                    )
                )

            normal = max(
                pyo.value(abs(enth_in)), pyo.value(abs(enth_out)), 1e-4
            )  # FIXME justify this number
            fraction_change = pyo.value((enth_out - enth_in) / normal)
            if abs(fraction_change) > 3e-3:
                raise RuntimeError(
                    f"Energy is not being conserved in {self.name}; "
                    f"fractional change {fraction_change}"
                )

    def recursive_scaling(self):
        gsf = iscale.get_scaling_factor

        def ssf(c, s):
            iscale.set_scaling_factor(c, s, overwrite=False)

        sgsf = iscale.set_and_get_scaling_factor

        def cst(c, s):
            iscale.constraint_scaling_transform(c, s, overwrite=False)

        sR = 1e-1  # Scaling factor for R
        # sD = 5e3 # Heuristic scaling factor for diffusion coefficient
        sD = 1e4
        sy_def = 10  # Mole frac comp scaling
        sh = 1e-2  # Heat xfer coeff
        # sh = 1
        sH = 1e-4  # Enthalpy/int energy
        sLx = sgsf(self.length_x, 1 / self.length_x.value)
        sLy = 1 / self.length_y[None].value

        for t in self.flowsheet().time:
            sT = sgsf(self.temperature_inlet[t], 1e-2)
            ssf(self.temperature_outlet[t], sT)
            sP = sgsf(self.pressure_inlet[t], 1e-4)

            s_flow_mol = sgsf(self.flow_mol_inlet[t], 1e3)
            sy_in = {}
            for j in self.component_list:
                sy_in[j] = sgsf(self.mole_frac_comp_inlet[t, j], sy_def)

            for iz in self.iznodes:
                # These should have been scaled by the cell-level method, so
                # notify the user if they're using a standalone channel
                # and forgot to scale these
                if not self.temperature_z[t, iz].is_reference():
                    gsf(self.temperature_z[t, iz], warning=True)
                gsf(self.heat_flux_x0[t, iz], warning=True)
                gsf(self.heat_flux_x1[t, iz], warning=True)

                s_flow_mol = sgsf(self.flow_mol[t, iz], s_flow_mol)
                sT = sgsf(self.temperature_z[t, iz], sT)
                sP = sgsf(self.pressure[t, iz], sP)
                cst(self.constant_pressure_eqn[t, iz], sP)

                cst(self.flow_mol_eqn[t, iz], s_flow_mol)
                sV = sR * sT / sP
                ssf(self.velocity[t, iz], sV * s_flow_mol / (sLx * sLy))

                sH = sgsf(self.enth_mol[t, iz], sH)
                cst(self.enth_mol_eqn[t, iz], sH)
                cst(self.energy_balance_eqn[t, iz], sH * s_flow_mol)

                if self.config.has_holdup:
                    sU = sgsf(self.int_energy_mol[t, iz], sH)
                    cst(self.int_energy_mol_eqn[t, iz], sU)

                    s_rho_U = sgsf(self.int_energy_density[t, iz], sU / sV)
                    cst(self.int_energy_density_eqn[t, iz], s_rho_U)

                sq0 = sgsf(self.heat_flux_x0[t, iz], 1e-2)
                cst(self.temperature_x0_eqn[t, iz], sq0)
                sq1 = sgsf(self.heat_flux_x1[t, iz], 1e-2)
                cst(self.temperature_x1_eqn[t, iz], sq1)
                sq = min(sq0, sq1)

                sDT = sgsf(self.temperature_deviation_x[t, iz], sq / sh)
                ssf(self.temperature_deviation_x0[t, iz], sDT)
                ssf(self.temperature_deviation_x1[t, iz], sDT)

                # Pointless other than making a record that this equation
                # is well-scaled by default
                cst(self.mole_frac_comp_eqn[t, iz], 1)

                for j in self.component_list:
                    # These should have been scaled by the cell-level method, so
                    # notify the user if they're using a standalone channel
                    # and forgot to scale these
                    if self.config.below_electrode:
                        gsf(self.material_flux_x1[t, iz, j], warning=True)
                    else:
                        gsf(self.material_flux_x0[t, iz, j], warning=True)

                    sy = sgsf(self.mole_frac_comp[t, iz, j], sy_in[j])
                    cst(self.material_balance_eqn[t, iz, j], s_flow_mol * sy)

                    ssf(self.conc_mol_comp[t, iz, j], sy * sP / (sR * sT))
                    cst(self.conc_mol_comp_eqn[t, iz, j], sy * sP)
                    # FIXME come back later to clean up
                    ssf(self.diff_eff_coeff[t, iz, j], 1e5)
                    cst(self.diff_eff_coeff_eqn[t, iz, j], 1e5)

                    if hasattr(self, "material_flux_x0_eqn"):
                        sXflux = gsf(
                            self.material_flux_x0[t, iz, j], default=1e-1, warning=True
                        )
                        cst(self.material_flux_x0_eqn[t, iz, j], sLx * sXflux / sD)
                        ssf(
                            self.conc_mol_comp_deviation_x0[t, iz, j], sLx * sXflux / sD
                        )
                    if hasattr(self, "material_flux_x1_eqn"):
                        sXflux = gsf(
                            self.material_flux_x1[t, iz, j], default=1e-1, warning=True
                        )
                        cst(self.material_flux_x1_eqn[t, iz, j], sLx * sXflux / sD)
                        ssf(
                            self.conc_mol_comp_deviation_x1[t, iz, j], sLx * sXflux / sD
                        )

            cst(self.temperature_outlet_eqn[t], sH)



#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES).
#
# Copyright (c) 2018-2024 by the software owners: The Regents of the
# University of California, through Lawrence Berkeley National Laboratory,
# National Technology & Engineering Solutions of Sandia, LLC, Carnegie Mellon
# University, West Virginia University Research Corporation, et al.
# All rights reserved.  Please see the files COPYRIGHT.md and LICENSE.md
# for full copyright and license information.
#################################################################################
import numpy as np
from copy import deepcopy

from ..geometry import Parallelepiped
from ..transform_func import ScaleFunc
from .unit_cell_lattice import UnitCell, UnitCellLattice
from ..tiling import CubicTiling
from ...util.util import ListHasPoint


class CubicLattice(UnitCellLattice):
    RefIAD = 1

    # === STANDARD CONSTRUCTOR
    def __init__(self, IAD):
        RefUnitCellShape = Parallelepiped.fromEdgesAndAngles(
            CubicLattice.RefIAD,
            CubicLattice.RefIAD,
            CubicLattice.RefIAD,
            np.pi / 2,
            np.pi / 2,
            np.pi / 2,
            np.array([0, 0, 0], dtype=float),
        )
        RefUnitCellTiling = CubicTiling(RefUnitCellShape)
        RefFracPositions = [np.array([0.0, 0.0, 0.0])]
        RefUnitCell = UnitCell(RefUnitCellTiling, RefFracPositions)
        UnitCellLattice.__init__(self, RefUnitCell)
        self._IAD = CubicLattice.RefIAD
        self.applyTransF(ScaleFunc(IAD / CubicLattice.RefIAD))
        self._NthNeighbors = [
            [
                np.array([1.0, 0.0, 0.0]),
                np.array([0.0, 1.0, 0.0]),
                np.array([0.0, 0.0, 1.0]),
                np.array([-1.0, 0.0, 0.0]),
                np.array([0.0, -1.0, 0.0]),
                np.array([0.0, 0.0, -1.0]),
            ]
        ]

    # === MANIPULATION METHODS
    def applyTransF(self, TransF):
        if isinstance(TransF, ScaleFunc):
            if TransF.isIsometric:
                self._IAD *= TransF.Scale[0]
            else:
                raise ValueError(
                    "CubicLattice applyTransF: Can only scale isometrically"
                )
        UnitCellLattice.applyTransF(self, TransF)

    # === PROPERTY EVALUATION METHODS
    # NOTE: This method is inherited from UnitCellLattice
    # def isOnLattice(self,P):

    def areNeighbors(self, P1, P2):
        return np.linalg.norm(P2 - P1) <= self.IAD

    def getNeighbors(self, P, layer=1):
        RefP = self._getConvertToReference(P)
        if layer > len(self._NthNeighbors):
            self._calculateNeighbors(layer)
        NBs = deepcopy(self._NthNeighbors[layer - 1])
        for NeighP in NBs:
            NeighP += RefP
            self._convertFromReference(NeighP)
        return NBs

    def _calculateNeighbors(self, layer):
        NList = [np.array([0, 0, 0], dtype=float)]
        for nb in self._NthNeighbors:
            NList.extend(nb)
        for _ in range(layer - len(self._NthNeighbors)):
            tmp = []
            for P in self._NthNeighbors[len(self._NthNeighbors) - 1]:
                for Q in self._NthNeighbors[0]:
                    N = P + Q
                    if not ListHasPoint(NList, N, 0.001 * CubicLattice.RefIAD):
                        tmp.append(N)
                        NList.append(N)
            self._NthNeighbors.append(tmp)

    # === BASIC QUERY METHODS
    @property
    def IAD(self):
        return self._IAD



#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES).
#
# Copyright (c) 2018-2024 by the software owners: The Regents of the
# University of California, through Lawrence Berkeley National Laboratory,
# National Technology & Engineering Solutions of Sandia, LLC, Carnegie Mellon
# University, West Virginia University Research Corporation, et al.
# All rights reserved.  Please see the files COPYRIGHT.md and LICENSE.md
# for full copyright and license information.
#################################################################################
"""
Tests for column model utitliy functions

Author: Alejandro Garciadiego, Andrew Lee
"""
import pytest
from types import MethodType

from pyomo.environ import ConcreteModel, Block, Var, Expression, Reference, Set, value
from pyomo.network import Port

from idaes.models_extra.column_models.util import make_phase_split

from idaes.core.util.exceptions import (
    PropertyPackageError,
    PropertyNotSupportedError,
    PropertyPackageError,
)

# Mark module as all unit tests
pytestmark = pytest.mark.unit


@pytest.fixture
def m():
    # Mock up something that looks like a tray model
    m = ConcreteModel()
    m.time = Set(initialize=[1, 2, 3])

    def flowsheet(m):
        return m

    m.flowsheet = MethodType(flowsheet, m)

    m.properties_out = Block(m.time)

    m.port = Port()

    return m


def test_make_phase_split_general(m):
    # Test for variables without special indices
    # Mock define_port_members method
    def define_port_members(m):
        return {"temperature": m.temperature, "indexed": m.indexed}

    # Add variables and define_port_members method to dummy state block
    for i in m.time:
        m.properties_out[i].temperature = Var()
        m.properties_out[i].indexed = Var(["a", "b"])

        m.properties_out[i].define_port_members = MethodType(
            define_port_members, m.properties_out[i]
        )

    # Call make_phase_split function
    make_phase_split(m, m.port)

    # Check for expected results
    for i in m.time:
        assert m.port.temperature[i] is m.properties_out[i].temperature
        for j in ["a", "b"]:
            assert m.port.indexed[i, j] is m.properties_out[i].indexed[j]


def test_make_phase_split_frac(m):
    # Mock define_port_members method
    def define_port_members(m):
        return {
            "flow_mass_phase": m.flow_mass_phase,
            "mass_frac_phase_comp": m.mass_frac_phase_comp,
        }

    def _liquid_set(m):
        return ["L", "L1"]

    def _vapor_set(m):
        return ["V", "V1"]

    m.components = Set(initialize=["a1", "b1"])
    m.phases = Set(initialize=["L", "V"])

    m.config.property_package = DummyData()
    m.config.property_package.component_list = m.components
    m.config.property_package.phase_list = m.phases

    m._liquid_set = MethodType(_liquid_set, m)
    m._vapor_set = MethodType(_vapor_set, m)

    # Add variables and define_port_members method to dummy state block
    for i in m.time:
        m.properties_out[i].flow_mass_phase = Var(m.phases)
        m.properties_out[i].mass_frac_phase_comp = Var(m.phases, m.components)

        m.properties_out[i].define_port_members = MethodType(
            define_port_members, m.properties_out[i]
        )

    # Call make_phase_split function
    make_phase_split(m, m.port)

    # Check for expected results
    for i in m.time:
        for p in m.config.property_package.phase_list:
            for j in m.config.property_package.component_list:
                assert (
                    m.port.flow_mass_phase[i, p, j] is m.e_flow_mass_phaseport[i, p, j]
                )
                assert value(m.port.flow_mass_phase[i, p, j]) == 1e-8
                assert (
                    m.port.mass_frac_phase_comp[i, p, j]
                    is m.properties_out[i].mass_frac_phase_comp[p, j]
                )


def test_make_phase_mole_frac(m):
    # Mock define_port_members method
    def define_port_members(m):
        return {
            "flow_mol_phase_comp": m.flow_mol_phase_comp,
            "mole_frac_comp": m.mole_frac_comp,
        }

    def _liquid_set(m):
        return ["L", "L1"]

    def _vapor_set(m):
        return ["V", "V1"]

    m.components = Set(initialize=["a1", "b1"])
    m.phases = Set(initialize=["L", "V"])

    m.config.property_package = DummyData()
    m.config.property_package.component_list = m.components
    m.config.property_package.phase_list = m.phases

    m._liquid_set = MethodType(_liquid_set, m)
    m._vapor_set = MethodType(_vapor_set, m)

    # Add variables and define_port_members method to dummy state block
    for i in m.time:
        m.properties_out[i].flow_mol_phase_comp = Var(m.phases, m.components)
        m.properties_out[i].mole_frac_comp = Var(m.components)

        m.properties_out[i].define_port_members = MethodType(
            define_port_members, m.properties_out[i]
        )

    # Call make_phase_split function
    make_phase_split(m, m.port, m.phases)

    # Check for expected results
    for i in m.time:
        for p in m.config.property_package.phase_list:
            for j in m.config.property_package.component_list:
                m.port.flow_mol_phase_comp[i, p, j] is m.e_flow_mol_phase_compport[
                    i, p, j
                ]
                assert value(m.port.flow_mol_phase_comp[i, p, j]) == 1e-8
                assert m.port.mole_frac_comp[i, j] is m.e_mole_frac_port[i, j]


def test_make_phase_split_mass_flow(m):
    # Mock define_port_members method
    def define_port_members(m):
        return {
            # "flow_mass": m.flow_mass,
            "mass_frac_comp": m.mass_frac_comp,
            "flow_mass_phase": m.flow_mass_phase,
            "mass_frac_phase_comp": m.mass_frac_phase_comp,
        }

    def _liquid_set(m):
        return ["a1", "b1"]

    def _vapor_set(m):
        return ["a1", "b1"]

    m.components = Set(initialize=["a1", "b1"])
    m.phases = Set(initialize=["L", "V"])

    m.config.property_package = DummyData()
    m.config.property_package.component_list = m.components
    m.config.property_package.phase_list = m.phases

    m._liquid_set = MethodType(_liquid_set, m)

    m._vapor_set = MethodType(_vapor_set, m)

    # Add variables and define_port_members method to dummy state block
    for i in m.time:
        # m.properties_out[i].flow_mass = Var()
        m.properties_out[i].mass_frac_comp = Var(m.components)
        m.properties_out[i].mass_frac_phase_comp = Var(m.phases, m.components)
        m.properties_out[i].flow_mass_phase = Var(m.phases)

        m.properties_out[i].define_port_members = MethodType(
            define_port_members, m.properties_out[i]
        )

    # Call make_phase_split function
    make_phase_split(m, m.port, m.phases, side_sf=1)

    # Check for expected results
    for i in m.time:
        for p in m.config.property_package.phase_list:
            for j in m.config.property_package.component_list:
                assert (
                    m.port.mass_frac_phase_comp[i, p, j]
                    is m._mass_frac_phase_comp_port_ref[i, p, j]
                )
                assert m.port.mass_frac_comp[i, j] is m.e_mole_frac_port[i, j]


def test_make_phase_split_mass(m):
    # Mock define_port_members method
    def define_port_members(m):
        return {
            # "flow_mass": m.flow_mass,
            "mass_frac_comp": m.mass_frac_comp,
            "flow_mass_phase_comp": m.flow_mass_phase_comp,
        }

    def _liquid_set(m):
        return ["a1", "b1"]

    def _vapor_set(m):
        return ["a1", "b1"]

    m.components = Set(initialize=["a1", "b1"])
    m.phases = Set(initialize=["L", "V"])

    m.config.property_package = DummyData()
    m.config.property_package.component_list = m.components
    m.config.property_package.phase_list = m.phases

    m._liquid_set = MethodType(_liquid_set, m)

    m._vapor_set = MethodType(_vapor_set, m)

    # Add variables and define_port_members method to dummy state block
    for i in m.time:
        # m.properties_out[i].flow_mass = Var()
        m.properties_out[i].mass_frac_comp = Var(m.components)
        m.properties_out[i].flow_mass_phase_comp = Var(m.phases, m.components)

        m.properties_out[i].define_port_members = MethodType(
            define_port_members, m.properties_out[i]
        )

    # Call make_phase_split function
    make_phase_split(m, m.port, m.phases, side_sf=1)

    # Check for expected results
    for i in m.time:
        for p in m.config.property_package.phase_list:
            for j in m.config.property_package.component_list:
                assert m.port.mass_frac_comp[i, j] is m.e_mole_frac_port[i, j]


def test_make_phase_split_enth(m):
    # Mock define_port_members method
    def define_port_members(m):
        return {
            "enth_mol_phase": m.enth_mol_phase,
            "enth_mol_phase_comp": m.enth_mol_phase_comp,
        }

    m.components = Set(initialize=["a1", "b1"])
    m.phases = Set(initialize=["L", "V"])

    m.config.property_package = DummyData()
    m.config.property_package.component_list = m.components
    m.config.property_package.phase_list = m.phases

    # Add variables and define_port_members method to dummy state block
    for i in m.time:
        m.properties_out[i].enth_mol_phase = Var(m.phases)
        m.properties_out[i].enth_mol_phase_comp = Var(m.phases, m.components)

        m.properties_out[i].define_port_members = MethodType(
            define_port_members, m.properties_out[i]
        )

    # Call make_phase_split function
    make_phase_split(m, m.port)

    # Check for expected results
    for i in m.time:
        for p in m.config.property_package.phase_list:
            assert m.port.enth_mol_phase[i, p] is m.properties_out[i].enth_mol_phase[p]
            for j in m.config.property_package.component_list:
                assert (
                    m.port.enth_mol_phase_comp[i, p, j]
                    is m.properties_out[i].enth_mol_phase_comp[p, j]
                )


def test_make_phase_split_enth_no_phase(m):
    # Mock define_port_members method
    def define_port_members(m):
        return {
            "enth_mass": m.enth_mass,
            "enth_mass_phase": m.enth_mass_phase,
        }

    # m.components = Set(initialize=["a1", "b1"])
    m.phases = Set(initialize=["L", "V"])

    m.config.property_package = DummyData()
    # m.config.property_package.component_list = m.components
    m.config.property_package.phase_list = m.phases

    # Add variables and define_port_members method to dummy state block
    for i in m.time:
        m.properties_out[i].enth_mass = Var()
        m.properties_out[i].enth_mass_phase = Var(m.phases)

        m.properties_out[i].define_port_members = MethodType(
            define_port_members, m.properties_out[i]
        )

    # Call make_phase_split function
    make_phase_split(m, m.port, m.phases)
    # Check for expected results
    for i in m.time:
        assert m.port.enth_mass[i] is m.e_enth_port[i]


def test_make_phase_non_mol_error(m):
    # Mock define_port_members method
    def define_port_members(m):
        return {
            "frac_comp": m.frac_comp,
        }

    m.components = Set(initialize=["a1", "b1"])
    m.phases = Set(initialize=["L", "V"])

    m.config.property_package = DummyData()
    m.config.property_package.component_list = m.components
    m.config.property_package.phase_list = m.phases

    # Add variables and define_port_members method to dummy state block
    for i in m.time:
        m.properties_out[i].frac_comp = Var(m.phases, m.components)

        m.properties_out[i].define_port_members = MethodType(
            define_port_members, m.properties_out[i]
        )

    # Check for expected results
    with pytest.raises(
        PropertyNotSupportedError,
        match="No mass frac or mole frac variables encountered "
        " while building ports. "
        "phase_frac as a state variable is not "
        "supported with distillation unit models.",
    ):
        make_phase_split(m, m.port)


def test_make_phase_mol_error(m):
    # Mock define_port_members method
    def define_port_members(m):
        return {
            "mole_frac_comp": m.mole_frac_comp,
        }

    m.components = Set(initialize=["a1", "b1"])
    m.phases = Set(initialize=["L", "V"])

    m.config.property_package = DummyData()
    m.config.property_package.component_list = m.components
    m.config.property_package.phase_list = m.phases

    # Add variables and define_port_members method to dummy state block
    for i in m.time:
        m.properties_out[i].mole_frac_comp = Var(m.phases, m.components)

        m.properties_out[i].define_port_members = MethodType(
            define_port_members, m.properties_out[i]
        )

    # Check for expected results
    with pytest.raises(
        PropertyNotSupportedError,
        match="No mole_frac_phase_comp or flow_mol_phase or"
        " flow_mol_phase_comp variables encountered "
        "while building ports. ",
    ):
        make_phase_split(m, m.port)


def test_make_phase_enth_error(m):
    # Mock define_port_members method
    def define_port_members(m):
        return {
            "enth_mol": m.enth_mol,
        }

    m.phases = Set(initialize=["L", "V"])

    m.config.property_package = DummyData()

    # Add variables and define_port_members method to dummy state block
    for i in m.time:
        m.properties_out[i].enth_mol = Var(m.phases)

        m.properties_out[i].define_port_members = MethodType(
            define_port_members, m.properties_out[i]
        )

    # Check for expected results
    with pytest.raises(
        PropertyPackageError,
        match="Enthalpy is indexed but the variable "
        "name does not reflect the presence of an index. "
        "Please follow the naming convention outlined "
        "in the documentation for state variables.",
    ):
        make_phase_split(m, m.port)


def test_make_phase_mass_error(m):
    # Mock define_port_members method
    def define_port_members(m):
        return {
            "mass_frac_comp": m.mass_frac_comp,
        }

    m.components = Set(initialize=["a1", "b1"])
    m.phases = Set(initialize=["L", "V"])

    m.config.property_package = DummyData()
    m.config.property_package.component_list = m.components
    m.config.property_package.phase_list = m.phases

    # Add variables and define_port_members method to dummy state block
    for i in m.time:
        m.properties_out[i].mass_frac_comp = Var(m.phases, m.components)

        m.properties_out[i].define_port_members = MethodType(
            define_port_members, m.properties_out[i]
        )

    # Check for expected results
    with pytest.raises(
        PropertyNotSupportedError,
        match="No mass_frac_phase_comp or flow_mass_phase or"
        " flow_mass_phase_comp variables encountered "
        "while building ports.",
    ):
        make_phase_split(m, m.port)


class DummyData:
    """
    Class containing necessary data to generate component list
    """

    def __init__(self):
        self.config = self.config()

    class config:
        def __init__(self):
            self.property_package = self.property_package()

        class property_package:
            def component_list(self, component_list):
                self.component_list = component_list

            def phase_list(self, phase_list):
                self.phase_list = phase_list



#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES).
#
# Copyright (c) 2018-2024 by the software owners: The Regents of the
# University of California, through Lawrence Berkeley National Laboratory,
# National Technology & Engineering Solutions of Sandia, LLC, Carnegie Mellon
# University, West Virginia University Research Corporation, et al.
# All rights reserved.  Please see the files COPYRIGHT.md and LICENSE.md
# for full copyright and license information.
#################################################################################
import pytest
import pyomo.environ as pyo
import idaes.core
from idaes.models_extra.power_generation.unit_models.helm import HelmSplitter
from idaes.models.properties import iapws95
from idaes.models.properties.general_helmholtz import helmholtz_available


@pytest.mark.skipif(not helmholtz_available(), reason="General Helmholtz not available")
@pytest.mark.component
def test_splitter():
    m = pyo.ConcreteModel()
    m.fs = idaes.core.FlowsheetBlock(dynamic=False)
    m.fs.properties = iapws95.Iapws95ParameterBlock()
    m.fs.unit = HelmSplitter(
        property_package=m.fs.properties, outlet_list=["o1", "o2", "o3"]
    )

    Fin = 1e4  # mol/s
    hin = 4000  # J/mol
    Pin = 101325  # Pa

    m.fs.unit.inlet.flow_mol[0].fix(Fin)
    m.fs.unit.inlet.enth_mol[0].fix(hin)
    m.fs.unit.inlet.pressure[0].fix(Pin)
    m.fs.unit.split_fraction[0, "o2"] = 0.3
    m.fs.unit.split_fraction[0, "o3"] = 0.2

    m.fs.unit.initialize()

    assert pyo.value(m.fs.unit.o1.flow_mol[0]) == pytest.approx(1e4 * 0.5, rel=1e-7)
    assert pyo.value(m.fs.unit.o2.flow_mol[0]) == pytest.approx(1e4 * 0.3, rel=1e-7)
    assert pyo.value(m.fs.unit.o3.flow_mol[0]) == pytest.approx(1e4 * 0.2, rel=1e-7)
    assert pyo.value(m.fs.unit.o1.pressure[0]) == pytest.approx(101325, rel=1e-7)
    assert pyo.value(m.fs.unit.o2.pressure[0]) == pytest.approx(101325, rel=1e-7)
    assert pyo.value(m.fs.unit.o3.pressure[0]) == pytest.approx(101325, rel=1e-7)
    assert pyo.value(m.fs.unit.o1.enth_mol[0]) == pytest.approx(4000, rel=1e-7)
    assert pyo.value(m.fs.unit.o2.enth_mol[0]) == pytest.approx(4000, rel=1e-7)
    assert pyo.value(m.fs.unit.o3.enth_mol[0]) == pytest.approx(4000, rel=1e-7)


@pytest.mark.skipif(not helmholtz_available(), reason="General Helmholtz not available")
@pytest.mark.unit
def test_get_stream_table_contents():
    m = pyo.ConcreteModel()
    m.fs = idaes.core.FlowsheetBlock(dynamic=False)
    m.fs.properties = iapws95.Iapws95ParameterBlock()
    m.fs.unit = HelmSplitter(
        property_package=m.fs.properties, outlet_list=["o1", "o2", "o3"]
    )

    stable = m.fs.unit._get_stream_table_contents()

    expected = {
        "Units": {
            "Mass Flow": getattr(pyo.units.pint_registry, "kg/s"),
            "Molar Flow": getattr(pyo.units.pint_registry, "mol/s"),
            "Molar Enthalpy": getattr(pyo.units.pint_registry, "J/mol"),
            "P": getattr(pyo.units.pint_registry, "Pa"),
            "T": getattr(pyo.units.pint_registry, "K"),
            "Vapor Fraction": getattr(pyo.units.pint_registry, "dimensionless"),
        },
        "inlet": {
            "Mass Flow": pytest.approx(0.01801527, rel=1e-5),
            "Molar Flow": pytest.approx(1.0, rel=1e-5),
            "Molar Enthalpy": pytest.approx(0.01102139, rel=1e-5),
            "P": pytest.approx(11032300, rel=1e-5),
            "T": pytest.approx(270.4877, rel=1e-5),
            "Vapor Fraction": pytest.approx(0.0, abs=1e-5),
        },
        "o1": {
            "Mass Flow": pytest.approx(0.01801527, rel=1e-5),
            "Molar Flow": pytest.approx(1.0, rel=1e-5),
            "Molar Enthalpy": pytest.approx(0.01102139, rel=1e-5),
            "P": pytest.approx(11032300, rel=1e-5),
            "T": pytest.approx(270.4877, rel=1e-5),
            "Vapor Fraction": pytest.approx(0.0, abs=1e-5),
        },
        "o2": {
            "Mass Flow": pytest.approx(0.01801527, rel=1e-5),
            "Molar Flow": pytest.approx(1.0, rel=1e-5),
            "Molar Enthalpy": pytest.approx(0.01102139, rel=1e-5),
            "P": pytest.approx(11032300, rel=1e-5),
            "T": pytest.approx(270.4877, rel=1e-5),
            "Vapor Fraction": pytest.approx(0.0, abs=1e-5),
        },
        "o3": {
            "Mass Flow": pytest.approx(0.01801527, rel=1e-5),
            "Molar Flow": pytest.approx(1.0, rel=1e-5),
            "Molar Enthalpy": pytest.approx(0.01102139, rel=1e-5),
            "P": pytest.approx(11032300, rel=1e-5),
            "T": pytest.approx(270.4877, rel=1e-5),
            "Vapor Fraction": pytest.approx(0.0, abs=1e-5),
        },
    }

    assert stable.to_dict() == expected



#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES).
#
# Copyright (c) 2018-2024 by the software owners: The Regents of the
# University of California, through Lawrence Berkeley National Laboratory,
# National Technology & Engineering Solutions of Sandia, LLC, Carnegie Mellon
# University, West Virginia University Research Corporation, et al.
# All rights reserved.  Please see the files COPYRIGHT.md and LICENSE.md
# for full copyright and license information.
#################################################################################

"""Tests for `MultiPeriodModel` class """

__author__ = "Radhakrishna Tumbalam Gooty"

import pytest
import matplotlib.pyplot as plt
import pyomo.environ as pyo
from idaes.core import FlowsheetBlock
from idaes.apps.grid_integration.multiperiod.multiperiod import MultiPeriodModel
from idaes.core.util.model_statistics import degrees_of_freedom
from idaes.core.util.exceptions import InitializationError
import idaes.logger as idaeslog


def build_flowsheet(m=None, a=None, b=None):
    """This function builds a dummy flowsheet"""
    if m is None:
        m = pyo.ConcreteModel()

    m.fs = FlowsheetBlock(dynamic=False)

    m.fs.x = pyo.Var(within=pyo.NonNegativeReals)
    m.fs.y = pyo.Var(within=pyo.NonNegativeReals)
    m.fs.con1 = pyo.Constraint(expr=m.fs.x + m.fs.y == 1)

    return m


def fix_dof_and_initialize(m):
    """This function fixes dof and initializes the dummy flowsheet"""
    m.fs.y.fix(0.5)
    m.fs.x.set_value(0.5)


def fix_dof_and_initialize_2(m):
    """This function fixes dof and initializes the dummy flowsheet"""
    m.fs.con2 = pyo.Constraint(expr=m.fs.x + m.fs.y == 2)


def unfix_dof(m):
    """This function unfixes dof for optimization"""
    m.fs.y.unfix()


def get_linking_variable_pairs(m1, m2):
    """This function returns pairs of linking variables"""
    return [(m1.fs.y, m2.fs.y)]


@pytest.fixture(scope="module")
def build_multi_period_model():
    m = MultiPeriodModel(
        n_time_points=5,
        process_model_func=build_flowsheet,
        linking_variable_func=get_linking_variable_pairs,
        use_stochastic_build=True,
        initialization_func=fix_dof_and_initialize,
        unfix_dof_func=unfix_dof,
    )

    return m


@pytest.fixture(scope="module")
def build_multi_period_stochastic_model():
    m = MultiPeriodModel(
        n_time_points=5,
        process_model_func=build_flowsheet,
        linking_variable_func=get_linking_variable_pairs,
        use_stochastic_build=True,
        set_scenarios=[1, 2, 3],
        initialization_func=fix_dof_and_initialize,
        unfix_dof_func=unfix_dof,
    )

    return m


@pytest.fixture(scope="module")
def build_multi_day_model():
    m = MultiPeriodModel(
        n_time_points=5,
        process_model_func=build_flowsheet,
        linking_variable_func=get_linking_variable_pairs,
        use_stochastic_build=True,
        set_days=["d1", "d2", "d3", "d4"],
        initialization_func=fix_dof_and_initialize,
        unfix_dof_func=unfix_dof,
    )

    return m


@pytest.fixture(scope="module")
def build_multi_day_stochastic_model():
    m = MultiPeriodModel(
        n_time_points=5,
        process_model_func=build_flowsheet,
        linking_variable_func=get_linking_variable_pairs,
        use_stochastic_build=True,
        set_days=["d1", "d2", "d3", "d4"],
        set_scenarios=[1, 2, 3],
        initialization_func=fix_dof_and_initialize,
        unfix_dof_func=unfix_dof,
    )

    return m


@pytest.fixture(scope="module")
def build_multi_year_model():
    m = MultiPeriodModel(
        n_time_points=5,
        process_model_func=build_flowsheet,
        linking_variable_func=get_linking_variable_pairs,
        use_stochastic_build=True,
        set_years=["y1", "y2", "y3"],
        initialization_func=fix_dof_and_initialize,
        unfix_dof_func=unfix_dof,
    )

    return m


@pytest.fixture(scope="module")
def build_multi_year_stochastic_model():
    m = MultiPeriodModel(
        n_time_points=5,
        process_model_func=build_flowsheet,
        linking_variable_func=get_linking_variable_pairs,
        use_stochastic_build=True,
        set_years=["y1", "y2", "y3"],
        set_scenarios=[1, 2, 3],
        initialization_func=fix_dof_and_initialize,
        unfix_dof_func=unfix_dof,
    )

    return m


@pytest.fixture(scope="module")
def build_multi_day_year_model():
    m = MultiPeriodModel(
        n_time_points=5,
        process_model_func=build_flowsheet,
        linking_variable_func=get_linking_variable_pairs,
        use_stochastic_build=True,
        set_days=["d1", "d2", "d3", "d4"],
        set_years=["y1", "y2", "y3"],
        initialization_func=fix_dof_and_initialize,
        unfix_dof_func=unfix_dof,
    )

    return m


@pytest.fixture(scope="module")
def build_multi_day_year_stochastic_model():
    m = MultiPeriodModel(
        n_time_points=5,
        process_model_func=build_flowsheet,
        linking_variable_func=get_linking_variable_pairs,
        use_stochastic_build=True,
        set_days=["d1", "d2", "d3", "d4"],
        set_years=["y1", "y2", "y3"],
        set_scenarios=[1, 2, 3],
        initialization_func=fix_dof_and_initialize,
        unfix_dof_func=unfix_dof,
    )

    return m


@pytest.fixture(scope="module")
def build_multi_period_model_manual():
    m = MultiPeriodModel(
        n_time_points=5,
        process_model_func=build_flowsheet,
        linking_variable_func=get_linking_variable_pairs,
        initialization_func=fix_dof_and_initialize,
        unfix_dof_func=unfix_dof,
    )
    m.build_multi_period_model()
    return m


@pytest.mark.unit
def test_lmp_plot(monkeypatch):
    lmp = [2, 3, 5, 6, 8, 6, 0, 4]
    lmp2 = {"Day1": lmp, "Day2": lmp}
    lmp3 = {
        "Day1": lmp,
        "Day2": lmp,
        "Day3": lmp,
        "Day4": lmp,
        "Day5": lmp,
        "Day6": lmp,
        "Day7": lmp,
    }
    time_vec = [i for i in range(len(lmp))]
    time_vec2 = {"Day1": time_vec, "Day2": time_vec}

    monkeypatch.setattr(plt, "show", lambda: None)

    MultiPeriodModel.plot_lmp_signal(lmp)
    MultiPeriodModel.plot_lmp_signal(lmp, time_vec, x_range=(0, 8), y_range=(0, 10))
    MultiPeriodModel.plot_lmp_signal(lmp2)
    MultiPeriodModel.plot_lmp_signal(lmp2, time_vec2)

    with pytest.raises(
        Exception,
        match=(
            f"Number of LMP signals provided exceeds six: the maximum "
            f"number of subplots the function supports."
        ),
    ):
        MultiPeriodModel.plot_lmp_signal(lmp3)


@pytest.mark.unit
def test_lmp_and_schedule_plot(monkeypatch):
    lmp = [2, 3, 5, 6, 8, 6, 0, 4]
    time_vec = [i for i in range(len(lmp))]

    schedule1 = {
        "var1": [2, 3, 4, 6, 5, 8, 2, 7],
        "var2": [2, 3, 4, 6, 5, 8, 2, 7],
    }

    schedule2 = {
        "var1": [2, 3, 4, 6, 5, 8, 2, 7],
        "var2": [2, 3, 4, 6, 5, 8, 2, 7],
        "var3": [2, 3, 4, 6, 5, 8, 2, 7],
        "var4": [2, 3, 4, 6, 5, 8, 2, 7],
        "var5": [2, 3, 4, 6, 5, 8, 2, 7],
    }

    monkeypatch.setattr(plt, "show", lambda: None)

    with pytest.raises(
        Exception,
        match=(
            f"Number of elements in schedule exceeds four: "
            f"the maximum number of subplots the function supports."
        ),
    ):
        MultiPeriodModel.plot_lmp_and_schedule(lmp, schedule2)

    MultiPeriodModel.plot_lmp_and_schedule(
        lmp, schedule1, color={1: "tab:blue", 2: "magenta"}
    )
    MultiPeriodModel.plot_lmp_and_schedule(
        lmp,
        schedule1,
        time_vec,
        x_range=(0, 8),
        lmp_range=(0, 9),
        y_label={"var1": "Power [MW]"},
        y_range={"var2": (0, 9)},
    )


@pytest.mark.unit
def test_multi_period_model(build_multi_period_model):
    m = build_multi_period_model

    assert hasattr(m, "period")
    assert len(m.period) == 5
    assert hasattr(m, "link_constraints")
    for i in [1, 2, 3, 4]:
        assert hasattr(m.link_constraints[i], "link_constraints")

    assert degrees_of_freedom(m) == 1


@pytest.mark.unit
def test_multi_period_stochastic_model(build_multi_period_stochastic_model):
    m = build_multi_period_stochastic_model

    assert not hasattr(m, "period")
    assert not hasattr(m, "link_constraints")
    assert hasattr(m, "scenario")
    assert len(m.scenario) == 3

    for i in [1, 2, 3]:
        assert hasattr(m.scenario[i], "period")
        assert hasattr(m.scenario[i], "link_constraints")
        assert len(m.scenario[i].period) == 5

        for j in [1, 2, 3, 4]:
            assert hasattr(m.scenario[i].link_constraints[j], "link_constraints")

    assert degrees_of_freedom(m) == 3  # num_scenarios


@pytest.mark.unit
def test_multi_day_model(build_multi_day_model):
    m = build_multi_day_model

    assert hasattr(m, "period")
    assert hasattr(m, "link_constraints")
    for d in ["d1", "d2", "d3", "d4"]:
        for t in [1, 2, 3, 4, 5]:
            assert (t, d) in m.period

            if t != 5:
                assert (t, d) in m.link_constraints
                assert hasattr(m.link_constraints[t, d], "link_constraints")

    assert degrees_of_freedom(m) == 4  # num_days


@pytest.mark.unit
def test_multi_day_stochastic_model(build_multi_day_stochastic_model):
    m = build_multi_day_stochastic_model

    assert not hasattr(m, "period")
    assert not hasattr(m, "link_constraints")
    assert hasattr(m, "scenario")

    for s in m.scenario:
        for d in ["d1", "d2", "d3", "d4"]:
            for t in [1, 2, 3, 4, 5]:
                assert (t, d) in m.scenario[s].period

                if t != 5:
                    assert (t, d) in m.scenario[s].link_constraints
                    assert hasattr(
                        m.scenario[s].link_constraints[t, d], "link_constraints"
                    )

    assert degrees_of_freedom(m) == 12  # num_days * num_scenarios


@pytest.mark.unit
def test_multi_year_model(build_multi_year_model):
    m = build_multi_year_model

    assert hasattr(m, "period")
    assert hasattr(m, "link_constraints")
    for y in ["y1", "y2", "y3"]:
        for t in [1, 2, 3, 4, 5]:
            assert (t, y) in m.period

            if t != 5:
                assert (t, y) in m.link_constraints
                assert hasattr(m.link_constraints[t, y], "link_constraints")

    assert degrees_of_freedom(m) == 3  # num_years


@pytest.mark.unit
def test_multi_year_stochastic_model(build_multi_year_stochastic_model):
    m = build_multi_year_stochastic_model

    assert not hasattr(m, "period")
    assert not hasattr(m, "link_constraints")
    assert hasattr(m, "scenario")

    for s in m.scenario:
        for y in ["y1", "y2", "y3"]:
            for t in [1, 2, 3, 4, 5]:
                assert (t, y) in m.scenario[s].period

                if t != 5:
                    assert (t, y) in m.scenario[s].link_constraints
                    assert hasattr(
                        m.scenario[s].link_constraints[t, y], "link_constraints"
                    )

    assert degrees_of_freedom(m) == 9  # num_years * num_scenarios


@pytest.mark.unit
def test_multi_day_year_model(build_multi_day_year_model):
    m = build_multi_day_year_model

    assert hasattr(m, "period")
    assert hasattr(m, "link_constraints")
    for y in ["y1", "y2", "y3"]:
        for d in ["d1", "d2", "d3", "d4"]:
            for t in [1, 2, 3, 4, 5]:
                assert (t, d, y) in m.period

                if t != 5:
                    assert (t, d, y) in m.link_constraints
                    assert hasattr(m.link_constraints[t, d, y], "link_constraints")

    assert degrees_of_freedom(m) == 12  # num_days * num_years


@pytest.mark.unit
def test_multi_day_year_stochastic_model(build_multi_day_year_stochastic_model):
    m = build_multi_day_year_stochastic_model

    assert not hasattr(m, "period")
    assert not hasattr(m, "link_constraints")
    assert hasattr(m, "scenario")

    for s in m.scenario:
        for y in ["y1", "y2", "y3"]:
            for d in ["d1", "d2", "d3", "d4"]:
                for t in [1, 2, 3, 4, 5]:
                    assert (t, d, y) in m.scenario[s].period

                    if t != 5:
                        assert (t, d, y) in m.scenario[s].link_constraints
                        assert hasattr(
                            m.scenario[s].link_constraints[t, d, y], "link_constraints"
                        )

    assert degrees_of_freedom(m) == 36  # num_days * num_years * num_scenarios


@pytest.mark.unit
def test_no_initialization(caplog):
    # # Cover warning associated with unfix_dof not provided
    with caplog.at_level(idaeslog.WARNING):
        m = MultiPeriodModel(
            n_time_points=2,
            process_model_func=build_flowsheet,
            linking_variable_func=get_linking_variable_pairs,
            use_stochastic_build=True,
            initialization_func=fix_dof_and_initialize,
        )

    assert (
        "unfix_dof_func argument is not provided. "
        "Returning the model without unfixing degrees of freedom." in caplog.text
    )

    # Cover warning associated with initialization_func not provided
    caplog.clear()
    with caplog.at_level(idaeslog.WARNING):
        m = MultiPeriodModel(
            n_time_points=2,
            process_model_func=build_flowsheet,
            linking_variable_func=get_linking_variable_pairs,
            use_stochastic_build=True,
        )

    assert (
        "initialization_func argument was not provided. Returning the multiperiod model without initialization."
        in caplog.text
    )

    # Cover warning associated with linking_variable_func not provided
    caplog.clear()
    with caplog.at_level(idaeslog.WARNING):
        m = MultiPeriodModel(
            n_time_points=2,
            process_model_func=build_flowsheet,
            linking_variable_func=None,
            use_stochastic_build=True,
        )

    assert (
        "linking_variable_func is not provided, so variables across"
        " time periods are not linked." in caplog.text
    )

    # Cover warning associated with periodic_variable_func provided
    caplog.clear()
    with caplog.at_level(idaeslog.WARNING):
        m = MultiPeriodModel(
            n_time_points=2,
            process_model_func=build_flowsheet,
            linking_variable_func=get_linking_variable_pairs,
            periodic_variable_func=get_linking_variable_pairs,
            use_stochastic_build=True,
            initialization_func=fix_dof_and_initialize,
            unfix_dof_func=unfix_dof,
        )

    assert (
        "A method is provided for get_periodic_variable_pairs. "
        "build_stochastic_multi_period method does not support periodic "
        "constraints, so the user needs to add them manually." in caplog.text
    )


@pytest.mark.unit
def test_data_kwargs(caplog, n_time_points=3):
    # Cover message associated with using model_data_kwargs
    with caplog.at_level(idaeslog.WARNING):
        m = MultiPeriodModel(
            n_time_points=n_time_points,
            process_model_func=build_flowsheet,
            linking_variable_func=get_linking_variable_pairs,
            initialization_func=fix_dof_and_initialize,
            unfix_dof_func=unfix_dof,
        )

        data = {
            0: {
                "a": 10,
                "b": 20,
            },
            1: {
                "a": 12,
                "b": 18,
            },
            2: {
                "a": 15,
                "b": 22,
            },
        }

        m.build_multi_period_model(model_data_kwargs=data)

        assert (
            f"model_data_kwargs argument is provided, so the flowsheet "
            f"options are different for different time instances. In this case, "
            f"the multiperiod model is returned without initialization." in caplog.text
        )


@pytest.mark.unit
def test_data_kwargs_fail(caplog, n_time_points=2):
    # Cover warning associated with len(model_data_kwargs) != n_time_points
    with caplog.at_level(idaeslog.WARNING):
        m = MultiPeriodModel(
            n_time_points=n_time_points,
            process_model_func=build_flowsheet,
            linking_variable_func=get_linking_variable_pairs,
            initialization_func=fix_dof_and_initialize,
            unfix_dof_func=unfix_dof,
        )

        data = {
            0: {
                "a": 10,
                "b": 20,
            },
            1: {
                "a": 12,
                "b": 18,
            },
            2: {
                "a": 15,
                "b": 22,
            },
        }

        m.build_multi_period_model(model_data_kwargs=data)

        assert (
            f"len(model_data_kwargs) != n_time_points.\n "
            f"len(model_data_kwargs) = {len(data)}\n"
            f"len(n_time_points) = {n_time_points}\n"
            f"Check input data for model_data_kwargs argument." in caplog.text
        )


@pytest.mark.unit
def test_initialization_fail():
    # Cover warning associated with the flowsheet failing to converge after initialization
    with pytest.raises(
        InitializationError,
        match=(
            f"Flowsheet did not converge after fixing the degrees of freedom. "
            f"To create the multi-period model without initialization, do not provide"
            f" initialization_func argument."
        ),
    ):
        m = MultiPeriodModel(
            n_time_points=5,
            process_model_func=build_flowsheet,
            linking_variable_func=get_linking_variable_pairs,
            use_stochastic_build=True,
            initialization_func=fix_dof_and_initialize_2,
            unfix_dof_func=unfix_dof,
        )


@pytest.mark.unit
def test_multi_period_model_manual(build_multi_period_model_manual):
    m = build_multi_period_model_manual

    assert hasattr(m, "blocks")
    assert len(m.blocks) == 5

    assert degrees_of_freedom(m) == 1



#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES).
#
# Copyright (c) 2018-2024 by the software owners: The Regents of the
# University of California, through Lawrence Berkeley National Laboratory,
# National Technology & Engineering Solutions of Sandia, LLC, Carnegie Mellon
# University, West Virginia University Research Corporation, et al.
# All rights reserved.  Please see the files COPYRIGHT.md and LICENSE.md
# for full copyright and license information.
#################################################################################
import pytest

from pyomo.environ import ConcreteModel, Expression, Set, Var, Constraint
from pyomo.util.check_units import assert_units_consistent

from idaes.core import (
    ControlVolume0DBlock,
    FlowsheetBlock,
    PhysicalParameterBlock,
    StateBlockData,
    MaterialBalanceType,
    EnergyBalanceType,
    MaterialFlowBasis,
)
from idaes.core.util.model_statistics import degrees_of_freedom
from idaes.core.solvers import get_solver
import idaes.core.util.scaling as iscale
from idaes.core.util.exceptions import InitializationError


from pyomo.core.base.var import VarData
from pyomo.core.base.param import ParamData
from pyomo.core.base.expression import ExpressionData
from idaes.core.initialization.initializer_base import (
    InitializationStatus,
)

_scalable = (VarData, ParamData, ExpressionData)

# -----------------------------------------------------------------------------
# Get default solver for testing
solver = get_solver("ipopt_v2")


# -----------------------------------------------------------------------------
class PropertyTestHarness(object):
    def configure_class(self, m):
        self.prop_pack = None
        self.param_args = {}
        self.prop_args = {}

        self.has_density_terms = True

        self.skip_initialization_raises_exception_test = False

        self.configure()

        # Need to attach these to m for later use
        m.has_density_terms = self.has_density_terms
        m.prop_args = self.prop_args
        m.skip_initialization_raises_exception_test = (
            self.skip_initialization_raises_exception_test
        )

    def configure(self):
        # Placeholder method to allow user to setup test harness
        pass

    @pytest.fixture(scope="class")
    def frame(self):
        m = ConcreteModel()
        self.configure_class(m)

        m.fs = FlowsheetBlock(dynamic=False)

        m.fs.params = self.prop_pack(**self.param_args)

        m.fs.props = m.fs.params.build_state_block(
            [1], defined_state=True, **m.prop_args
        )

        return m

    @pytest.mark.unit
    def test_param_block(self, frame):
        if not isinstance(frame.fs.params, PhysicalParameterBlock):
            raise TypeError(
                "Parameter Block does not inherit from PhysicalParameterBlock"
            )

    @pytest.mark.unit
    def test_component_list(self, frame):
        if not isinstance(frame.fs.params.component_list, Set):
            raise TypeError(
                "Parameter block does not have a component_list, or "
                "component_list is not a Pyomo Set."
            )

    @pytest.mark.unit
    def test_phase_list(self, frame):
        if not isinstance(frame.fs.params.phase_list, Set):
            raise TypeError(
                "Parameter block does not have a phase_list, or "
                "phase_list is not a Pyomo Set."
            )

    @pytest.mark.unit
    def test_state_block_class(self, frame):
        if not hasattr(frame.fs.params, "state_block_class"):
            raise AttributeError("Parameter block does not specify state_block_class.")

    @pytest.mark.unit
    def test_properties_meta_data(self, frame):
        if frame.fs.params.get_metadata().properties is None:
            raise AttributeError(
                "Parameter block has not specified properties metadata."
            )

    @pytest.mark.unit
    def test_default_units_meta_data(self, frame):
        if frame.fs.params.get_metadata().default_units is None:
            raise AttributeError(
                "Parameter block has not specified default_units metadata."
            )

    @pytest.mark.unit
    def test_state_block_construction(self, frame):
        if not isinstance(frame.fs.props[1], StateBlockData):
            raise TypeError("State Block does not inherit from StateBlockData.")

    @pytest.mark.unit
    def test_state_block_temperature(self, frame):
        if not isinstance(frame.fs.props[1].temperature, (Var, Expression)):
            raise AttributeError(
                "State Block does not contain temperature, or temperature is "
                "not a Pyomo Var or Expression."
            )

    @pytest.mark.unit
    def test_state_block_pressure(self, frame):
        if not isinstance(frame.fs.props[1].pressure, (Var, Expression)):
            raise AttributeError(
                "State Block does not contain pressure, or pressure is "
                "not a Pyomo Var or Expression."
            )

    @pytest.mark.unit
    def test_state_block_mole_frac_phase_comp(self, frame):
        if not isinstance(frame.fs.props[1].mole_frac_phase_comp, (Var, Expression)):
            raise AttributeError(
                "State Block does not contain mole_frac_phase_comp, or "
                "mole_frac_phase_comp is not a Pyomo Var or Expression."
            )
        err = False
        if len(frame.fs.props[1].mole_frac_phase_comp) != (
            len(frame.fs.params._phase_component_set)
        ):
            err = True

        for k in frame.fs.props[1].mole_frac_phase_comp:
            if (
                k[0] not in frame.fs.params.phase_list
                or k[1] not in frame.fs.params.component_list
            ):
                err = True

        if err:
            raise ValueError(
                "mole_frac_phase_comp is not indexed by phase and component."
            )

    @pytest.mark.unit
    def test_get_material_flow_terms(self, frame):
        try:
            for p, j in frame.fs.params._phase_component_set:
                term = frame.fs.props[1].get_material_flow_terms(p, j)
                # Assert that the term can be assigned a scale factor
                assert isinstance(term, _scalable)
        except KeyError:
            raise KeyError(
                "get_material_flow_terms method is not indexed by phase and "
                "component."
            )
        except AttributeError:
            raise AttributeError(
                "State block has not implemented get_material_flow_terms " "method."
            )

    @pytest.mark.unit
    def test_get_enthalpy_flow_terms(self, frame):
        try:
            for p in frame.fs.params.phase_list:
                term = frame.fs.props[1].get_enthalpy_flow_terms(p)
                # Assert that the term can be assigned a scale factor
                assert isinstance(term, _scalable)
        except KeyError:
            raise KeyError("get_enthalpy_flow_terms method is not indexed by phase.")
        except AttributeError:
            raise AttributeError(
                "State block has not implemented get_enthalpy_flow_terms " "method."
            )

    @pytest.mark.unit
    def test_get_material_density_terms(self, frame):
        if frame.has_density_terms:
            try:
                for p in frame.fs.params.phase_list:
                    for j in frame.fs.params.component_list:
                        term = frame.fs.props[1].get_material_density_terms(p, j)
                        # Assert that the term can be assigned a scale factor
                        assert isinstance(term, _scalable)
            except KeyError:
                raise KeyError(
                    "get_material_density_terms method is not indexed by phase"
                    " and component."
                )
            except AttributeError:
                raise AttributeError(
                    "State block has not implemented "
                    "get_material_density_terms method."
                )

    @pytest.mark.unit
    def test_get_energy_density_terms(self, frame):
        if frame.has_density_terms:
            try:
                for p in frame.fs.params.phase_list:
                    term = frame.fs.props[1].get_energy_density_terms(p)
                    assert isinstance(term, _scalable)
            except KeyError:
                raise KeyError(
                    "get_enthalpy_density_terms method is not indexed by " "phase."
                )
            except AttributeError:
                raise AttributeError(
                    "State block has not implemented "
                    "get_enthalpy_density_terms method."
                )

    @pytest.mark.unit
    def test_default_material_balance_type(self, frame):
        if frame.fs.props[1].default_material_balance_type() not in MaterialBalanceType:
            raise ValueError("Invalid value for default_material_balance_type.")

    @pytest.mark.unit
    def test_default_energy_balance_type(self, frame):
        if frame.fs.props[1].default_energy_balance_type() not in EnergyBalanceType:
            raise ValueError("Invalid value for default_energy_balance_type.")

    @pytest.mark.unit
    def test_get_material_flow_basis(self, frame):
        if frame.fs.props[1].get_material_flow_basis() not in MaterialFlowBasis:
            raise ValueError("Invalid value for get_material_flow_basis.")

    @pytest.mark.unit
    def test_define_state_vars(self, frame):
        try:
            sv = frame.fs.props[1].define_state_vars()
        except AttributeError:
            raise AttributeError(
                "State block has not implemented define_state_vars method."
            )

        if not isinstance(sv, dict):
            raise TypeError("define_state_vars must return a dictionary of Vars.")

        for v in sv.values():
            if not isinstance(v, Var):
                raise TypeError(
                    "Invalid entry in define_state_Vars, {}. All members must "
                    "be Pyomo Vars.".format(v)
                )

    @pytest.mark.component
    def test_unit_consistency(self, frame):
        assert_units_consistent(frame)

    @pytest.mark.component
    def test_initialize(self, frame):
        frame._init_dof = degrees_of_freedom(frame.fs.props[1])

        if not hasattr(frame.fs.props, "initialize") or not callable(
            frame.fs.props.initialize
        ):
            raise AttributeError(
                "State Block has not implemented an initialize method."
            )

        frame._flags = frame.fs.props.initialize(hold_state=True)

        if degrees_of_freedom(frame.fs.props[1]) != 0:
            # PYLINT-TODO
            # pylint: disable-next=broad-exception-raised
            raise Exception(
                "initialize did not result in a State Block with 0 "
                "degrees of freedom."
            )

    @pytest.mark.component
    def test_release_state(self, frame):
        if not hasattr(frame.fs.props, "release_state") or not callable(
            frame.fs.props.release_state
        ):
            raise AttributeError(
                "State Block has not implemented a release_state method."
            )

        frame.fs.props.release_state(frame._flags)

        if degrees_of_freedom(frame.fs.props[1]) != frame._init_dof:
            # PYLINT-TODO
            # pylint: disable-next=broad-exception-raised
            raise Exception(
                "release state did not restore State Block to original "
                "degrees of freedom."
            )

    @pytest.mark.component
    def test_initialize_failure(self, frame):
        if not frame.skip_initialization_raises_exception_test:
            for n, v in frame.fs.props[1].define_state_vars().items():
                for i in v:
                    frame.fs.props[1].add_component(
                        "_init_test_" + str(v) + str(i), Constraint(expr=v[i] <= 0)
                    )

            with pytest.raises(
                InitializationError,
                match="fs.props failed to initialize successfully. Please "
                "check the output logs for more information.",
            ):
                frame.fs.props.initialize()

            for n, v in frame.fs.props[1].define_state_vars().items():
                for i in v:
                    frame.fs.props[1].del_component("_init_test_" + str(v) + str(i))

    @pytest.mark.component
    def test_CV_integration(self, frame):
        # Phase equilibrium does not make sense for single phase systems, so check
        # and only set to True if more than one phase present.
        if len(frame.fs.params.phase_list) > 1:
            has_pe = True
        else:
            has_pe = False

        frame.fs.cv = ControlVolume0DBlock(property_package=frame.fs.params)

        frame.fs.cv.add_geometry()

        frame.fs.cv.add_state_blocks(has_phase_equilibrium=has_pe)

        frame.fs.cv.add_material_balances(has_phase_equilibrium=has_pe)

        frame.fs.cv.add_energy_balances()

        frame.fs.cv.add_momentum_balances()

    @pytest.mark.component
    def test_default_scaling_factors(self, frame):
        # check that the calculate_scaling_factors method successfully copies
        # the default scaling factors to the scaling suffixes.  If there are
        # no default scaling factors, this should pass
        iscale.calculate_scaling_factors(frame)  # this also ensure, it doesn't except
        for v in frame.fs.props[1].component_data_objects(
            (Constraint, Var, Expression), descend_into=False
        ):
            name = v.getname().split("[")[0]
            index = v.index()
            print(v)
            assert (
                iscale.get_scaling_factor(v)
                == frame.fs.props[1].config.parameters.get_default_scaling(name, index)
            ) or (
                frame.fs.props[1].config.parameters.get_default_scaling(name, index)
                is None
            )

    @pytest.mark.component
    def test_default_initializer(self):
        m = ConcreteModel()
        self.configure_class(m)

        m.fs = FlowsheetBlock(dynamic=False)

        m.fs.params = self.prop_pack(**self.param_args)

        m.fs.props = m.fs.params.build_state_block(
            [1], defined_state=True, **m.prop_args
        )

        # Check that a default initializer is defined
        initializer = m.fs.props.default_initializer
        assert initializer is not None

        # Most cases will probably have specified the class for the initializer, but
        # in case an instance was provided check to see if it is callable.
        if callable(initializer):
            initializer = initializer()

        # Check that we get a valid solution with the default initializer
        assert initializer.initialize(m.fs.props) == InitializationStatus.Ok
